/*############################################################################
# Keith R. Jackson, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################*/
%module ftpControlc
%{
#undef HAVE_LIMITS_H
#undef SIZEOF_OFF_T 
#undef socklen_t
#include "globus_common.h"
#include "globus_ftp_control.h"
#include "pyGlobus-threading.h"
%}

%include ./src/pyGlobus-typemaps.i

%types(callbackStruct*);
%types(callbackStruct4*);
%types(union globus_ftp_control_command_u*);
%types(globus_ftp_control_auth_info_t*);
%types(globus_byte_t*);
%types(globus_ftp_control_response_t *);
%types(globus_ftp_control_handle_s *);

typedef int globus_bool_t;


typedef enum globus_ftp_control_dcau_mode_e
{
    GLOBUS_FTP_CONTROL_DCAU_NONE = 'N',
    GLOBUS_FTP_CONTROL_DCAU_SELF = 'A',
    GLOBUS_FTP_CONTROL_DCAU_SUBJECT = 'S',
    GLOBUS_FTP_CONTROL_DCAU_DEFAULT
} globus_ftp_control_dcau_mode_t;

typedef enum
{
    GLOBUS_FTP_CONTROL_PROTECTION_CLEAR = 'C',
    GLOBUS_FTP_CONTROL_PROTECTION_SAFE = 'S',
    GLOBUS_FTP_CONTROL_PROTECTION_CONFIDENTIAL = 'E',
    GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE = 'P'
} globus_ftp_control_protection_t;

typedef enum globus_ftp_control_type_e
{
    GLOBUS_FTP_CONTROL_TYPE_NONE,
    GLOBUS_FTP_CONTROL_TYPE_ASCII = 'A',
    GLOBUS_FTP_CONTROL_TYPE_EBCDIC = 'E',
    GLOBUS_FTP_CONTROL_TYPE_IMAGE = 'I',
    GLOBUS_FTP_CONTROL_TYPE_LOCAL = 'L'
} globus_ftp_control_type_t;

typedef enum globus_ftp_control_mode_e
{
    GLOBUS_FTP_CONTROL_MODE_NONE, 
    GLOBUS_FTP_CONTROL_MODE_STREAM = 'S',
    GLOBUS_FTP_CONTROL_MODE_BLOCK = 'B', 
    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK = 'E',
    GLOBUS_FTP_CONTROL_MODE_COMPRESSED = 'C'
} globus_ftp_control_mode_t;

typedef enum globus_ftp_control_structure_e
{
    GLOBUS_FTP_CONTROL_STRUCTURE_NONE,
    GLOBUS_FTP_CONTROL_STRUCTURE_FILE,
    GLOBUS_FTP_CONTROL_STRUCTURE_PAGE,
    GLOBUS_FTP_CONTROL_STRUCTURE_RECORD
} globus_ftp_control_structure_t;

typedef enum globus_ftp_control_parallelism_mode_e
{
    GLOBUS_FTP_CONTROL_PARALLELISM_NONE,
    GLOBUS_FTP_CONTROL_PARALLELISM_FIXED,
} globus_ftp_control_parallelism_mode_t;

typedef enum globus_ftp_control_response_class_e
{
    GLOBUS_FTP_UNKNOWN_REPLY,
    GLOBUS_FTP_POSITIVE_PRELIMINARY_REPLY,
    GLOBUS_FTP_POSITIVE_COMPLETION_REPLY,
    GLOBUS_FTP_POSITIVE_INTERMEDIATE_REPLY,
    GLOBUS_FTP_TRANSIENT_NEGATIVE_COMPLETION_REPLY,
    GLOBUS_FTP_PERMANENT_NEGATIVE_COMPLETION_REPLY
} globus_ftp_control_response_class_t;

typedef enum globus_ftp_control_tcpbuffer_mode_e
{
    /** Don't change the TCP buffer/window size from the system default */
    GLOBUS_FTP_CONTROL_TCPBUFFER_DEFAULT,
    
    /** Set the TCP buffer/window size to a fixed value */
    GLOBUS_FTP_CONTROL_TCPBUFFER_FIXED,

    /** Automatically set the TCP buffer/window size */
    GLOBUS_FTP_CONTROL_TCPBUFFER_AUTOMATIC
} globus_ftp_control_tcpbuffer_mode_t;

typedef enum globus_ftp_data_connection_state_e
{
    GLOBUS_FTP_DATA_STATE_NONE, /* dc_handle has no references */
    GLOBUS_FTP_DATA_STATE_PASV, /* in local pasv mode */
    GLOBUS_FTP_DATA_STATE_PORT, /* in local port mode */
    GLOBUS_FTP_DATA_STATE_SPOR, /* in local spor mode */
    GLOBUS_FTP_DATA_STATE_CONNECT_READ, /* connected for reading */
    GLOBUS_FTP_DATA_STATE_CONNECT_WRITE, /* connected for writing */
    GLOBUS_FTP_DATA_STATE_CLOSING, /* closing all connections */
    GLOBUS_FTP_DATA_STATE_EOF, /* user has received eof */

    GLOBUS_FTP_DATA_STATE_SEND_EOF, /* not used for state at all */
} globus_ftp_data_connection_state_t;

typedef enum
{
    GLOBUS_FTP_CONTROL_UNCONNECTED,
    GLOBUS_FTP_CONTROL_CONNECTED,
    GLOBUS_FTP_CONTROL_CLOSING
} globus_ftp_cc_state_t;

typedef enum globus_ftp_control_command_code_e
{
    GLOBUS_FTP_CONTROL_COMMAND_OPTS,
    GLOBUS_FTP_CONTROL_COMMAND_AUTH,
    GLOBUS_FTP_CONTROL_COMMAND_ADAT,
    GLOBUS_FTP_CONTROL_COMMAND_SPAS,
    GLOBUS_FTP_CONTROL_COMMAND_SPOR,
    GLOBUS_FTP_CONTROL_COMMAND_PORT,
    GLOBUS_FTP_CONTROL_COMMAND_PASV,
    GLOBUS_FTP_CONTROL_COMMAND_SITE,
    GLOBUS_FTP_CONTROL_COMMAND_TYPE,
    GLOBUS_FTP_CONTROL_COMMAND_DELE,
    GLOBUS_FTP_CONTROL_COMMAND_FEAT,
    GLOBUS_FTP_CONTROL_COMMAND_ERET,
    GLOBUS_FTP_CONTROL_COMMAND_ESTO,
    GLOBUS_FTP_CONTROL_COMMAND_RMD,
    GLOBUS_FTP_CONTROL_COMMAND_MKD,
    GLOBUS_FTP_CONTROL_COMMAND_PWD,
    GLOBUS_FTP_CONTROL_COMMAND_CWD,
    GLOBUS_FTP_CONTROL_COMMAND_CDUP,
    GLOBUS_FTP_CONTROL_COMMAND_NLST,
    GLOBUS_FTP_CONTROL_COMMAND_HELP,
    GLOBUS_FTP_CONTROL_COMMAND_STAT,
    GLOBUS_FTP_CONTROL_COMMAND_NOOP,
    GLOBUS_FTP_CONTROL_COMMAND_SYST,
    GLOBUS_FTP_CONTROL_COMMAND_STOU,
    GLOBUS_FTP_CONTROL_COMMAND_QUIT,
    GLOBUS_FTP_CONTROL_COMMAND_REIN,
    GLOBUS_FTP_CONTROL_COMMAND_ABOR,
    GLOBUS_FTP_CONTROL_COMMAND_ALLO,
    GLOBUS_FTP_CONTROL_COMMAND_MODE,
    GLOBUS_FTP_CONTROL_COMMAND_STRU,
    GLOBUS_FTP_CONTROL_COMMAND_ACCT,
    GLOBUS_FTP_CONTROL_COMMAND_PASS,
    GLOBUS_FTP_CONTROL_COMMAND_USER,
    GLOBUS_FTP_CONTROL_COMMAND_SMNT,
    GLOBUS_FTP_CONTROL_COMMAND_LIST,
    GLOBUS_FTP_CONTROL_COMMAND_RETR,
    GLOBUS_FTP_CONTROL_COMMAND_REST,
    GLOBUS_FTP_CONTROL_COMMAND_SBUF,
    GLOBUS_FTP_CONTROL_COMMAND_SIZE,
    GLOBUS_FTP_CONTROL_COMMAND_STOR,
    GLOBUS_FTP_CONTROL_COMMAND_APPE,
    GLOBUS_FTP_CONTROL_COMMAND_RNFR,
    GLOBUS_FTP_CONTROL_COMMAND_RNTO,
    GLOBUS_FTP_CONTROL_COMMAND_UNKNOWN
} globus_ftp_control_command_code_t;

%constant GLOBUS_FTP_CONTROL_AUTH_REQ_NONE               =   1;
%constant GLOBUS_FTP_CONTROL_AUTH_REQ_GSSAPI	    =  2;
%constant GLOBUS_FTP_CONTROL_AUTH_REQ_USER	    =  4;
%constant GLOBUS_FTP_CONTROL_AUTH_REQ_PASS	    =  8;
%constant GLOBUS_FTP_CONTROL_AUTH_REQ_ACCT	    = 16;

typedef enum globus_ftp_control_type_option_e
{
    GLOBUS_FTP_CONTROL_TYPE_NO_OPTION,
    GLOBUS_FTP_CONTROL_TYPE_OPTION_N,
    GLOBUS_FTP_CONTROL_TYPE_OPTION_T,
    GLOBUS_FTP_CONTROL_TYPE_OPTION_C
} globus_ftp_control_type_option_t;

typedef enum globus_ftp_control_auth_type_e
{
    GLOBUS_FTP_CONTROL_AUTH_GSSAPI,
    GLOBUS_FTP_CONTROL_AUTH_UNKNOWN
} globus_ftp_control_auth_type_t;

/* fix up the command struct with a few accessors */


%init %{
    PyEval_InitThreads();
    python_interp = PyThreadState_Get()->interp;
    init_globus_thread_state_key(&python_thread_state_key);
%}

%{
static void python_thread_state_destroy_cb(void *val);
static globus_thread_key_t python_thread_state_key;
static PyInterpreterState *python_interp;

PyObject * check_error(globus_object_t *  error){
    PyObject *tuple, *err;

    if(!(tuple = PyTuple_New(2))) {
    	PyErr_SetString(PyExc_RuntimeError,"PythonCallback, PyTuple_New");	
    }
    else{ 
	if (error) {  /* SAME AS IF(ERROR==GLOBUS_SUCCESS)*/
            char *tmpstr;
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(error);
            PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
            globus_libc_free(tmpstr);
        }
	else{
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
            Py_INCREF(Py_None);
            PyTuple_SET_ITEM(tuple, 1, Py_None);
	}
    }	
    
    if ((err =  PyErr_Occurred()) != NULL)
    {
	PyErr_Print();
    }
    return tuple;
}


static void PythonCompleteAuthCallback( 
	void *                                   user_data,
    struct globus_ftp_control_handle_s *     handle,
    globus_object_t *                        error,
    globus_ftp_control_auth_info_t *         auth_result) 
{
    PyObject *func, *arglist, *result, *arg, *ptr, *ptr2, *tuple;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    int rslt=0;

    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));
    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_handle_t, 0);
    ptr2 = SWIG_NewPointerObj((void*)auth_result, 
			SWIGTYPE_p_globus_ftp_control_auth_info_t, 0);

    tuple = check_error(error);

    arglist = Py_BuildValue("(OOOO)", arg, ptr, tuple, ptr2);
    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
	PyErr_Print();
    }
    Py_DECREF(arglist);
    Py_XDECREF(result);
    Py_DECREF(ptr);
    Py_DECREF(ptr2);
    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
}   

static void PythonCompleteCommandCallback(
    void *                                   user_data,
    struct globus_ftp_control_handle_s *     handle,
    globus_object_t *                        error,
    union globus_ftp_control_command_u *     command)

{
    PyObject *func, *arglist, *result, *arg, *ptr, *ptr2, *tuple;

    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));
    ptr2 = SWIG_NewPointerObj((void*)command, 
			      SWIGTYPE_p_globus_ftp_control_command_u, 0);

    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_handle_t, 0);

    tuple = check_error(error);

    arglist = Py_BuildValue("(OOOO)", arg, ptr, tuple, ptr2);

    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
      PyErr_Print();
    }
    Py_DECREF(ptr);
    Py_DECREF(ptr2);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
}

static void PythonCompleteServerCallback(void* user_data,
    globus_ftp_control_server_t* 			handle,
    globus_object_t* 						error)
{
    PyObject *func, *arglist, *result, *arg, *ptr, *tuple;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));
    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_server_t, 0);

    tuple = check_error(error);
    
    arglist = Py_BuildValue("(OOO)", arg, ptr, tuple);
    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
      PyErr_Print();
    }
    Py_DECREF(ptr);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
}                       

static void PythonCompleteControlCallback(
	void* 							user_data,
	globus_ftp_control_handle_t* 	handle,
	globus_object_t* 				error) 
{
    PyObject *func, *arglist, *result, *arg, *ptr, *tuple;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));

    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_handle_t, 0);

    tuple = check_error(error);
    arglist = Py_BuildValue("(OOO)", arg, ptr, tuple);

    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
      PyErr_Print();
    }
    Py_DECREF(arglist);
    Py_XDECREF(result);
    Py_DECREF(ptr);

    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
}   

static void PythonCompleteDataCallback(
	void* 							user_data,
	globus_ftp_control_handle_t* 	handle,
	globus_object_t* 				error,
	globus_byte_t* 					buffer,
	globus_size_t 					length,
	globus_off_t 					offset,
	globus_bool_t 					eof) 
{
    PyObject *func, *arglist, *result, *arg, *buf, *oSet, *ptr, *ptr2, *oldBuf, *tuple;
    callbackStruct4* cbStruct;
    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));
    cbStruct = (callbackStruct4*)user_data;
    func = cbStruct->callback;
    arg = cbStruct->arg;
    oldBuf = cbStruct->buffer;

    tuple = check_error(error);

    if (sizeof(globus_off_t) == 8) {
        /* XXX, This assumes globus_off_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for
           different types for size_t */
        oSet = PyLong_FromLongLong(offset);
    }
    else {
        oSet = PyInt_FromLong(offset);
    }
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_handle_t, 0);
    ptr2 = SWIG_NewPointerObj((void*)buffer, SWIGTYPE_p_globus_byte_t, 0);
    buf = PyBuffer_FromReadWriteMemory((void*)buffer, length * sizeof(globus_byte_t));
    arglist = Py_BuildValue("(OOOOiOiO)", arg, ptr, buf, ptr2, length,
                            oSet, (int)eof, tuple);
    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
      PyErr_Print();
    }
    Py_XDECREF(oldBuf);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    
    /* XXX: jrl adding calls to clean up memory here!  */
    Py_DECREF(oSet);
    Py_DECREF(buf);
    Py_DECREF(ptr);
    Py_DECREF(ptr2);

    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
}   

static void PythonCompleteDataConnectCallback(
    void*                           user_data,
    globus_ftp_control_handle_t*    handle,
	unsigned int					stripe_ndx,
	globus_bool_t					reused,
    globus_object_t*                error)
{
    PyObject *func, *arglist, *result, *arg, *ptr, *tuple;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));
    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_handle_t, 0);

    tuple = check_error(error);

    arglist = Py_BuildValue("(OOiiO)", arg, ptr, stripe_ndx, reused, tuple);
    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
      PyErr_Print();
    }
    Py_DECREF(ptr);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
} 

static void PythonResponseCallback( void * user_data, globus_ftp_control_handle_t * handle,globus_object_t * error, globus_ftp_control_response_t *ftp_response)
{
    PyObject * func, *arglist, *result, *arg, *ptr1, *ptr2, *tuple;
    callbackStruct* cbStruct = (callbackStruct*)user_data;
    PyEval_AcquireThread(get_python_thread_state(python_thread_state_key, python_interp));
    func = cbStruct->callback;
    arg = cbStruct->arg;
    ptr1 = SWIG_NewPointerObj((void*)handle, 
			     SWIGTYPE_p_globus_ftp_control_handle_t, 0);
    ptr2 = SWIG_NewPointerObj((void*)ftp_response, SWIGTYPE_p_globus_ftp_control_response_t, 0);

    tuple = check_error(error);

    arglist = Py_BuildValue("(OOOO)", arg, ptr1, tuple, ptr2);
    result = PyEval_CallObject(func, arglist);
    if (result == NULL) {
      PyErr_Print();
    }
    Py_DECREF(ptr1);
    Py_DECREF(ptr2);
    Py_DECREF(arglist);
    Py_XDECREF(result);
    PyEval_ReleaseThread(get_python_thread_state(python_thread_state_key, python_interp));
    return;
}

%}

%inline %{

char *raw_command( union globus_ftp_control_command_u *  command)
{
	return (command->base.raw_command);
}

int code( union globus_ftp_control_command_u *  command)
{
	return (command->base.code);
}

char *operation( union globus_ftp_control_command_u *  command)
{
	char *op;
	int l;

	/* worst case */ 
	l = strlen(command->base.raw_command);	
	if ((op = calloc(l,sizeof(char))) == NULL)
		return NULL;
	sscanf(command->base.raw_command, "%s %*\n", op);
	return (op);
}


globus_module_descriptor_t*
get_module(void) {
    return GLOBUS_FTP_CONTROL_MODULE;
}

globus_ftp_control_layout_t*
layout_init(void) {
    return (globus_ftp_control_layout_t*)globus_malloc(sizeof(globus_ftp_control_layout_t));
}

void
layout_destroy(globus_ftp_control_layout_t* layout) {
    globus_free(layout);
}

globus_ftp_control_dcau_t*
dcau_init(void) {
    return (globus_ftp_control_dcau_t*)globus_malloc(sizeof(globus_ftp_control_dcau_t));
}

void
dcau_destroy(globus_ftp_control_dcau_t* dcau) {
    globus_free(dcau);
}

int
dcau_set_mode(globus_ftp_control_dcau_t* dcau, char mode) {
    globus_ftp_control_dcau_mode_t md;
    if (mode == 'N')
	md = GLOBUS_FTP_CONTROL_DCAU_NONE;
    else if (mode == 'A')
	md = GLOBUS_FTP_CONTROL_DCAU_SELF;
    else if (mode == 'S')
	md = GLOBUS_FTP_CONTROL_DCAU_SUBJECT;
    else 
	md = GLOBUS_FTP_CONTROL_DCAU_DEFAULT;

    dcau->mode = md;
    return GLOBUS_SUCCESS;
}

int 
dcau_set_subject(globus_ftp_control_dcau_t* dcau, char* subject) {
    dcau->subject.subject = subject;
    return GLOBUS_SUCCESS;
}

int
layout_set_striping_mode(globus_ftp_control_layout_t* layout, int mode) {
    globus_ftp_control_striping_mode_t md;
    if (mode == 0) 
        md = GLOBUS_FTP_CONTROL_STRIPING_NONE;
    else if (mode == 1)
        md = GLOBUS_FTP_CONTROL_STRIPING_PARTITIONED;
    else if (mode == 2)
        md = GLOBUS_FTP_CONTROL_STRIPING_BLOCKED_ROUND_ROBIN;
    else
        return -42;
    layout->mode = md;
    return GLOBUS_SUCCESS;
}

void
layout_set_round_robin_block_size(globus_ftp_control_layout_t* layout,
                                  PyObject* obj) {
    globus_size_t blockSize;
    if ((sizeof(globus_size_t) == 8) && (PyLong_Check(obj))) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
        blockSize = (globus_size_t)PyLong_AsLongLong(obj);
    }
    else {
        blockSize = (globus_size_t)PyInt_AsLong(obj);
    }
    layout->round_robin.block_size = blockSize;
    return;
}

void
layout_set_partition_block_size(globus_ftp_control_layout_t* layout,
                                PyObject* obj) {
    globus_size_t blockSize;
    if ((sizeof(globus_size_t) == 8) && (PyLong_Check(obj))) {
        /* XXX, This assumes size_t is a long long.  This code will
           probably have to be ifdef'd by platform to account for 
           different types for size_t */
        blockSize = (globus_size_t)PyLong_AsLongLong(obj);
    }
    else {
        blockSize = (globus_size_t)PyInt_AsLong(obj);
    }
    layout->partitioned.size = blockSize;
    return;    
}

PyObject *
parallelism_init(void) {
    
   PyObject *ptr;
   globus_ftp_control_parallelism_t * par; 
   par = (globus_ftp_control_parallelism_t*)globus_malloc(sizeof(globus_ftp_control_parallelism_t));
   ptr = SWIG_NewPointerObj((void *) par, SWIGTYPE_p_globus_ftp_control_parallelism_t, 0);
   return ptr;

}

void
parallelism_destroy(globus_ftp_control_parallelism_t* par) {
    globus_free(par);
}

int
parallelism_set_mode(globus_ftp_control_parallelism_t* par, int mode) {
    globus_ftp_control_parallelism_mode_t md; 
    if (mode == 0) 
    {
        md = GLOBUS_FTP_CONTROL_PARALLELISM_NONE;
    }
    else if (mode == 1)
    {
	md = GLOBUS_FTP_CONTROL_PARALLELISM_FIXED; 
    }
    else
    {
	return -42;
    }
    par->mode = md;
    return GLOBUS_SUCCESS;
}

int
parallelism_set_size( globus_ftp_control_parallelism_t* par, int size) {
	par->fixed.size = size;
	return GLOBUS_SUCCESS;
}

globus_ftp_control_tcpbuffer_t*
tcpbuffer_init(void) {
    return (globus_ftp_control_tcpbuffer_t*)globus_malloc(sizeof(globus_ftp_control_tcpbuffer_t));
}

void
tcpbuffer_destroy(globus_ftp_control_tcpbuffer_t* buffer) {
    globus_free(buffer);
}

void
tcpbuffer_set_automatic(globus_ftp_control_tcpbuffer_t* handle,
                        unsigned int initial, unsigned int minimum,
                        unsigned int maximum) {
    handle->mode = GLOBUS_FTP_CONTROL_TCPBUFFER_AUTOMATIC;
    handle->automatic.initial_size = initial;
    handle->automatic.minimum_size = minimum;
    handle->automatic.maximum_size = maximum;
    return;
}

void
tcpbuffer_set_fixed(globus_ftp_control_tcpbuffer_t* handle,
                    unsigned int size) {
    handle->mode = GLOBUS_FTP_CONTROL_TCPBUFFER_FIXED;
    handle->fixed.size = size;
    return;
}

void
tcpbuffer_set_default(globus_ftp_control_tcpbuffer_t* handle) {
    handle->mode = GLOBUS_FTP_CONTROL_TCPBUFFER_DEFAULT;
    return;
}

PyObject*
handle_init(void) {
    globus_ftp_control_handle_t* handle;
    globus_result_t ret;
    char* tmpstr;
    PyObject* tuple, *ptr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_init: PyTuple_New()");
        return NULL;
    }
    handle = (globus_ftp_control_handle_t*)globus_malloc(sizeof(globus_ftp_control_handle_t));
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_handle_init(handle);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        globus_libc_free(handle);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)handle, 
				 SWIGTYPE_p_globus_ftp_control_handle_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
handle_destroy(globus_ftp_control_handle_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "handle_destroy: PyTuple_New()");
        return NULL;
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_handle_destroy(handle);
    Py_END_ALLOW_THREADS
    globus_free(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

/* jrl */
/* This is the Server class of stuff in the Control lib wrapper */
PyObject*
server_handle_init(void) {
    globus_ftp_control_server_t* handle;
    globus_result_t ret;
    char* tmpstr;
    PyObject *tuple, *ptr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError,"server_handle_init: PyTuple_New()");
        return NULL;
    }

    handle = (globus_ftp_control_server_t *)globus_malloc(sizeof(globus_ftp_control_server_t));
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_server_handle_init(handle);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        globus_libc_free(handle);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        ptr = SWIG_NewPointerObj((void*)handle, 
				 SWIGTYPE_p_globus_ftp_control_server_t, 0);
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
server_handle_destroy(globus_ftp_control_server_t* handle) {
    globus_result_t ret;
    PyObject* tuple;
    char* tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "server_handle_destroy: PyTuple_New()");
        return NULL;
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_server_handle_destroy(handle);
    Py_END_ALLOW_THREADS
    globus_free(handle);
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
    }
    else {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
server_accept(globus_ftp_control_server_t* serv_handle,
	globus_ftp_control_handle_t* handle, 
	PyObject* pyfunc, PyObject* arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "server_accept: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_server_accept(serv_handle, handle,
                                PythonCompleteControlCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
server_authenticate( globus_ftp_control_handle_t* handle,
	int auth_req,
	PyObject* pyfunc, PyObject* arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "server_authenticate: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_server_authenticate(handle, 
				auth_req,
                PythonCompleteAuthCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
server_listen( globus_ftp_control_server_t* serv_handle,
	unsigned short port, 
	PyObject* pyfunc, PyObject* arg) 
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    unsigned short assigned_port = port;
    callbackStruct* cbStruct;

    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "server_listen: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_server_listen(serv_handle, 
						&assigned_port,
                        PythonCompleteServerCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
         PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("i", -1));
        
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
        PyTuple_SET_ITEM(tuple, 2, Py_BuildValue("i", assigned_port));
    }
    
    return tuple;
}

PyObject*
server_stop( globus_ftp_control_server_t* serv_handle,
	PyObject* pyfunc, PyObject* arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "server_stop: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_server_stop(serv_handle, 
                                PythonCompleteServerCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

/* misc stuff -- should there be a data class? local class? */

PyObject*
data_connect_write( 
	globus_ftp_control_handle_t* handle,
	PyObject* pyfunc, 
	PyObject* arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "data_connect_write: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_data_connect_write(handle,
                                PythonCompleteDataConnectCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
data_write( globus_ftp_control_handle_t* handle,
	PyObject*				pyBuffer,
	PyObject*				offset,
	globus_bool_t      	     eof,
	PyObject* 				pyfunc, 
	PyObject* 				arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_off_t converted_offset;
    globus_byte_t* buf;
    callbackStruct4* cbStruct;
    int sizeNew;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "data_write: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct4*)globus_malloc(sizeof(callbackStruct4));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    Py_XINCREF(pyBuffer);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    cbStruct->buffer = pyBuffer;

    if ((sizeof(globus_off_t) == 8) && (PyLong_Check(offset))) {
        /* XXX. this assumes off_t is a long long.  It will
           probably be necessary to ifdef this code for various
           platforms.  This works on 64bit Solaris 7, but I don't
           know about others. */
        converted_offset = (globus_off_t)PyLong_AsLongLong(offset);
    }
    else {
        converted_offset = (globus_off_t)PyInt_AsLong(offset);
    }
     PyString_AsStringAndSize(pyBuffer, (char**)&buf, &sizeNew);
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_data_write(handle, buf, sizeNew, 
					converted_offset, eof,
					PythonCompleteDataCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
	
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct4, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
local_pasv( globus_ftp_control_handle_t* handle) 
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_control_host_port_t * address;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "local_pasv: PyTuple_New()");
        return NULL;
    }
    address = (globus_ftp_control_host_port_t *)globus_malloc(sizeof(globus_ftp_control_host_port_t));
    address->host[0] = 0;
    address->host[1] = 0;
    address->host[2] = 0;
    address->host[3] = 0;
    address->port = 0;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_local_pasv(handle, address);
    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", ret));
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
    }
    else {
        ptr = SWIG_NewPointerObj((void *)address, SWIGTYPE_p_globus_ftp_control_host_port_t, 0);
	PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    /*
    printf("address->host[0]: %d\n", address->host[0]);
    printf("address->host[1]: %d\n", address->host[1]);
    printf("address->host[2]: %d\n", address->host[2]);
    printf("address->host[3]: %d\n", address->host[3]);
    printf("address->port: %d\n", address->port);
    */
    return tuple;
}

PyObject*
local_port( globus_ftp_control_handle_t* handle, 
	globus_ftp_control_host_port_t *            address)
{
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "local_port: PyTuple_New()");
        return NULL;
    }
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_local_port(handle, address);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        
    }
    else {
		PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        Py_INCREF(Py_None);
        PyTuple_SET_ITEM(tuple, 1, Py_None);
    }
    return tuple;
}

PyObject*
read_commands( globus_ftp_control_handle_t* handle,
	PyObject* pyfunc, PyObject* arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "read_commands: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_read_commands(handle, 
                                PythonCompleteCommandCallback, (void*)cbStruct);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}

PyObject*
send_response( globus_ftp_control_handle_t* handle,
	const char * 			respspec,
    PyObject* pyfunc, PyObject* arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "send_response: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;

    Py_BEGIN_ALLOW_THREADS
    ret = globus_ftp_control_send_response(handle,
								respspec,
                                PythonCompleteControlCallback, 
								(void*)cbStruct,NULL);
    Py_END_ALLOW_THREADS
    if (ret != GLOBUS_SUCCESS) {
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
        tmpstr = globus_object_printable_to_string(globus_error_get(ret));
        PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
        
        globus_libc_free(cbStruct);
    }
    else {
        ptr = SWIG_NewPointerObj((void*)cbStruct, 
				 SWIGTYPE_p_callbackStruct, 0);
        PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
        PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
}                

/***Matt mucks everything up after this ***/
/**Data registration functions **/
PyObject *
create_data_info(globus_ftp_control_handle_t* handle,globus_byte_t * buffer,globus_size_t length, globus_off_t offset, globus_bool_t eof, PyObject* pyfunc, PyObject *arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr, *ptr2;
    char* tmpstr;
    callbackStruct* cbStruct;
    globus_ftp_control_data_write_info_t * data_info;
    
    if (!(tuple=PyTuple_New(3)))
    {
        PyErr_SetString(PyExc_RuntimeError, "create_data_info: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    data_info = (globus_ftp_control_data_write_info_t *)globus_malloc(sizeof(globus_ftp_control_data_write_info_t));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
     Py_BEGIN_ALLOW_THREADS
     ret = globus_ftp_control_create_data_info(handle, data_info, buffer, length, offset, eof,PythonCompleteDataCallback, cbStruct);
     Py_END_ALLOW_THREADS
	 
	 if (ret != GLOBUS_SUCCESS){
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	     tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	     PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	     
	     globus_libc_free(cbStruct);
	 }
         else {
	     ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	     ptr2 = SWIG_NewPointerObj((void *)data_info, SWIGTYPE_p_globus_ftp_control_data_write_info_t, 0);
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	     PyTuple_SET_ITEM(tuple, 1, ptr);
	 }   
	     
	     return tuple;
}

PyObject *
release_data_info(globus_ftp_control_handle_t* handle,globus_ftp_control_data_write_info_t * data_info)
    {
	 globus_result_t ret;
	 char * tmpstr;
	 PyObject * tuple;
	 if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "release_data_info: PyTuple_New()");
        return NULL;
    }
	Py_BEGIN_ALLOW_THREADS
	    ret = globus_ftp_control_release_data_info(handle, data_info);
	Py_END_ALLOW_THREADS
	if (ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
         else {
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	     PyTuple_SET_ITEM(tuple, 1,  Py_BuildValue("i", 0) );
	 }   
	     
	return tuple;
	   

    }

PyObject *
data_write_stripe(globus_ftp_control_handle_t* handle, PyObject * pyBuffer, globus_off_t offset, 
		  globus_bool_t eof, int stripe_ndx,  PyObject* pyfunc, PyObject *arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct4* cbStruct;
    globus_byte_t *buf;
    int sizeNew;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "data_write_stripe: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct4*)globus_malloc(sizeof(callbackStruct4));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    Py_XINCREF(pyBuffer);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    cbStruct->buffer = pyBuffer;

    PyString_AsStringAndSize(pyBuffer, (char**)&buf, &sizeNew);
   
     Py_BEGIN_ALLOW_THREADS
	ret = globus_ftp_control_data_write_stripe(handle, buf, sizeNew, offset, eof, stripe_ndx,PythonCompleteDataCallback, cbStruct);
     Py_END_ALLOW_THREADS
	 if (ret != GLOBUS_SUCCESS){
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	     tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	     PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
         else {
	     ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct4, 0);
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	     PyTuple_SET_ITEM(tuple, 1, ptr);
	 }   
     
     return tuple;
}

PyObject *
x_data_write_stripe(globus_ftp_control_handle_t* handle,PyObject *pyBuffer, globus_off_t offset, globus_bool_t eof, int stripe_ndx, globus_ftp_control_data_write_info_t * data_info)
{
    globus_result_t ret;
    PyObject *tuple;
    char* tmpstr;
    globus_byte_t *buf;
    int sizeNew;

    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "x_data_write_stripe: PyTuple_New()");
        return NULL;
    }
     PyString_AsStringAndSize(pyBuffer, (char**)&buf, &sizeNew);
     Py_BEGIN_ALLOW_THREADS
	 
	 ret =  globus_X_ftp_control_data_write_stripe(handle, buf, sizeNew, offset, eof, stripe_ndx,data_info);
     Py_END_ALLOW_THREADS
	 if (ret != GLOBUS_SUCCESS){
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	     tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	     PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
         else {
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	     PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("i", 0));
	 }   
	     
     return tuple;
}
/***globus_ftp_control_client****/

PyObject *
control_connect(globus_ftp_control_handle_t* handle, char * host, unsigned port, PyObject * pyfunc, PyObject *arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct; 
    if (!(tuple=PyTuple_New(2)))
    {
        PyErr_SetString(PyExc_RuntimeError, "control_connect: PyTuple_New()");
        return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
	ret = globus_ftp_control_connect(handle, host, port,PythonResponseCallback, cbStruct);
    Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0); 
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;
}

PyObject *
response_destroy(globus_ftp_control_response_t * response)
{
    globus_result_t ret;
    PyObject * pyRet;
    char * tmpstr;
    ret = globus_ftp_control_response_destroy(response);
    if(ret != GLOBUS_SUCCESS){
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	pyRet = Py_BuildValue("s", tmpstr);
    }
    else {
	pyRet = Py_BuildValue("s", "NONE");
    }
    return pyRet;

}

PyObject *
response_copy(globus_ftp_control_response_t *src){
    globus_result_t ret;
    PyObject * tuple, *ptr;
    char * tmpstr;
    globus_ftp_control_response_t *dest;
    dest = (globus_ftp_control_response_t *)globus_malloc(sizeof(globus_ftp_control_response_t));
    if (!(tuple=PyTuple_New(2)))
	{
	    PyErr_SetString(PyExc_RuntimeError, "response_copy: PyTuple_New()");
	    return NULL;
	}

    ret = globus_ftp_control_response_copy(src, dest);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
       ptr = SWIG_NewPointerObj((void *)dest, SWIGTYPE_p_globus_ftp_control_response_t, 0);
       PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
       PyTuple_SET_ITEM(tuple, 1, ptr);
    }
    return tuple;
} 

  
PyObject *
authenticate(globus_ftp_control_handle_t *handle, globus_ftp_control_auth_info_t * auth_info, globus_bool_t use_auth, PyObject * pyfunc, PyObject *arg)
    {
	globus_result_t ret;
	PyObject *tuple, *ptr;
	char* tmpstr;
	callbackStruct* cbStruct;

	if (!(tuple=PyTuple_New(2)))
	{
	    PyErr_SetString(PyExc_RuntimeError, "authenticate: PyTuple_New()");
	    return NULL;
	}
	cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
	Py_XINCREF(pyfunc);
	Py_XINCREF(arg);
	cbStruct->callback = pyfunc;
	cbStruct->arg = arg;
	Py_BEGIN_ALLOW_THREADS
	    ret = globus_ftp_control_authenticate(handle, auth_info, use_auth, PythonResponseCallback, cbStruct);
	
	Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	     ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	     PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	     PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;
}
/*Variable length arguements not support on >cc -V
cc: Sun WorkShop 6 update 2 C 5.3 2001/05/15
*/
PyObject *
send_command(globus_ftp_control_handle_t *handle,const char * cmdspec,PyObject * pyfunc, PyObject * arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
    
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "send_command:: PyTuple_New()");
	return NULL;
    }
    cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
    Py_XINCREF(pyfunc);
    Py_XINCREF(arg);
    cbStruct->callback = pyfunc;
    cbStruct->arg = arg;
    Py_BEGIN_ALLOW_THREADS
	ret = globus_ftp_control_send_command(handle, cmdspec, PythonResponseCallback, cbStruct);
    Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;

}

PyObject *
py_abort(globus_ftp_control_handle_t *handle,PyObject * pyfunc, PyObject * arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
     if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "py_abort:: PyTuple_New()");
	return NULL;
    }
     cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
     Py_XINCREF(pyfunc);
     Py_XINCREF(arg);
     cbStruct->callback = pyfunc;
     cbStruct->arg = arg;
     Py_BEGIN_ALLOW_THREADS
	 ret = globus_ftp_control_abort(handle, PythonResponseCallback, cbStruct);
     Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;

}

PyObject *
py_quit(globus_ftp_control_handle_t *handle,PyObject * pyfunc, PyObject * arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
     if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "py_quit: PyTuple_New()");
	return NULL;
    }
     cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
     Py_XINCREF(pyfunc);
     Py_XINCREF(arg);
     cbStruct->callback = pyfunc;
     cbStruct->arg = arg;
     Py_BEGIN_ALLOW_THREADS
	 ret = globus_ftp_control_quit(handle, PythonResponseCallback, cbStruct);
     Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;

}

PyObject *
force_close(globus_ftp_control_handle_t *handle,PyObject * pyfunc, PyObject * arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
     if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "force_close: PyTuple_New()");
	return NULL;
    }
     cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
     Py_XINCREF(pyfunc);
     Py_XINCREF(arg);
     cbStruct->callback = pyfunc;
     cbStruct->arg = arg;
     Py_BEGIN_ALLOW_THREADS
	 ret = globus_ftp_control_force_close(handle, PythonResponseCallback, cbStruct);
     Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;

}

PyObject *
auth_info_init(gss_cred_id_t credential_handle, globus_bool_t encrypt, char * user, char * password, char * account, char * subject)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    globus_ftp_control_auth_info_t * auth_info;
    auth_info = (globus_ftp_control_auth_info_t *)globus_malloc(sizeof(globus_ftp_control_auth_info_t));
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "auth_info_init:: PyTuple_New()");
	return NULL;
    }
     ret = globus_ftp_control_auth_info_init(auth_info,credential_handle,encrypt,user,password,account, subject);
     if(ret != GLOBUS_SUCCESS){
	 PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	 tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	 PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	 
     }
     else{
	  PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	  ptr = SWIG_NewPointerObj((void *)auth_info, SWIGTYPE_p_globus_ftp_control_auth_info_t, 0);
	  PyTuple_SET_ITEM(tuple, 1, ptr);
     }
     
     return tuple;
}
void
auth_info_destroy(globus_ftp_control_auth_info_t * auth_info)
{
    globus_free(auth_info);
}

int
auth_info_compare(globus_ftp_control_auth_info_t * auth_info_1, globus_ftp_control_auth_info_t *auth_info_2)
{
    int ret;
    ret = globus_ftp_control_auth_info_compare(auth_info_1, auth_info_2);
    return ret;
}
/*
  PyObject *
  auth_info_get_user(globus_ftp_control_auth_info_t * auth_info)
  {
  PyObject * user;
  printf("user: %s\n", auth_info->user);
  user = PyString_FromString(auth_info->user);
  return user;
  }
  void
  auth_info_examine(globus_ftp_control_auth_info_t * auth_info)
  {
  printf("In examine auth_info\n");
  printf("command code: %d\n", auth_info->prev_cmd);
  printf("authenticated: %d\n", auth_info->authenticated);
  printf("auth_gssapi_subject: %s\n", auth_info->auth_gssapi_subject);
  printf("encrypt: %d\n", auth_info->encrypt);
  printf("user: %s\n", auth_info->user);
  printf("password: %s\n", auth_info->password);
  printf("account: %s\n", auth_info->account);
  }
  void 
  set_hack_authenticate(globus_ftp_control_auth_info_t * auth_info, int auth)
  {
  auth_info->authenticated = auth;
  }

  void set_hack_encrypt(globus_ftp_control_auth_info_t * auth_info, int enc)
  {
  auth_info->encrypt = enc; 
  }

  void set_hack_encrypt_handle(globus_ftp_control_handle_t * handle, int enc)
  {
  handle->cc_handle.auth_info.encrypt = enc;
  }
  
  PyObject *
  command_decode(globus_ftp_control_handle_t * handle, char * cmd_string)
  {
  globus_result_t ret;
  char * real_cmd = GLOBUS_NULL;
  char * tmpstr;
  PyObject * tuple;
  
    
  if (!(tuple=PyTuple_New(2)))
  {
  PyErr_SetString(PyExc_RuntimeError, "command_decode:: PyTuple_New()");
  return NULL;
  }
  ret = (globus_result_t)globus_i_ftp_control_decode_command(cmd_string, &real_cmd, &(handle->cc_handle.auth_info));
  PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
  
  if (ret !=GLOBUS_SUCCESS)
  {
  tmpstr = globus_object_printable_to_string(globus_error_get(ret));
  PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
  globus_free(tmpstr);
	}
	else
	{
    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", real_cmd));
    }
    }
*/

PyObject *
data_connect_read(globus_ftp_control_handle_t * handle, PyObject * pyfunc, PyObject * arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
     if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "data_connect_read: PyTuple_New()");
	return NULL;
    }
     cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
     Py_XINCREF(pyfunc);
     Py_XINCREF(arg);
     cbStruct->callback = pyfunc;
     cbStruct->arg = arg;
     Py_BEGIN_ALLOW_THREADS
	 ret = globus_ftp_control_data_connect_read(handle, PythonCompleteDataConnectCallback,cbStruct);
     Py_END_ALLOW_THREADS
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", ret)); 
     if(ret != GLOBUS_SUCCESS){
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
	    globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple; 
}
PyObject *
data_add_channels(globus_ftp_control_handle_t * handle, unsigned int num_channels, unsigned int stripe_ndx)
{
    globus_result_t ret;
    PyObject *retString;
    char * tmpstr;
   
    ret = globus_ftp_control_data_add_channels(handle, num_channels, stripe_ndx);
    if(ret != GLOBUS_SUCCESS){

	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	retString =  Py_BuildValue("s", tmpstr);
	
    }
    else {
	retString = Py_BuildValue("s", "NONE");
    }   
    
    return retString; 
}

PyObject *
data_send_eof(globus_ftp_control_handle_t * handle, int count[], int array_size, globus_bool_t eof_message, PyObject * pyfunc, PyObject * arg)
{
 globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct* cbStruct;
     if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "data_send_eof: PyTuple_New()");
	return NULL;
    }
     cbStruct = (callbackStruct*)globus_malloc(sizeof(callbackStruct));
     Py_XINCREF(pyfunc);
     Py_XINCREF(arg);
     cbStruct->callback = pyfunc;
     cbStruct->arg = arg;
     Py_BEGIN_ALLOW_THREADS
	 ret = globus_ftp_control_data_send_eof(handle,count, array_size, eof_message, PythonCompleteControlCallback, cbStruct);
     Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct, 0);
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
    return tuple;
}

PyObject *
data_remove_channels(globus_ftp_control_handle_t * handle, unsigned int num_channels, unsigned int stripe_ndx)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "data_remove_channels: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_data_remove_channels(handle, num_channels, stripe_ndx);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", "NONE"));
    }   
    
    return tuple; 
}

PyObject *
data_query_channels(globus_ftp_control_handle_t * handle, unsigned int stripe_ndx)
{
    globus_result_t ret;
    unsigned int * num_channels;
    PyObject *tuple, *num_chans;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "data_query_channels: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_data_query_channels(handle, num_channels, stripe_ndx);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	num_chans = PyLong_FromUnsignedLong(*num_channels);
	PyTuple_SET_ITEM(tuple, 1, num_chans);
    }   
    
    return tuple; 
}

PyObject *
data_get_total_data_channels(globus_ftp_control_handle_t * handle, unsigned int stripe_ndx)
{
    globus_result_t ret;
    PyObject *tuple, *pyNumChannels;
    char * tmpstr;
    unsigned int *num_channels;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "get_total_data_channels: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_data_get_total_data_channels(handle, num_channels, stripe_ndx);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	pyNumChannels = PyLong_FromUnsignedLong(*num_channels);
	PyTuple_SET_ITEM(tuple, 1, pyNumChannels);
    }   
    
    return tuple; 
}
PyObject *
local_send_eof(globus_ftp_control_handle_t * handle,globus_bool_t send_eof)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_send_eof: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_send_eof(handle, send_eof);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", "NONE"));
    }   
    
    return tuple; 

}
PyObject *
local_parallelism(globus_ftp_control_handle_t * handle,globus_ftp_control_parallelism_t * parallelism)
{
   
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_parallelism: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_parallelism(handle, parallelism);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;  

}
PyObject *
local_spas(globus_ftp_control_handle_t * handle, globus_ftp_control_host_port_t addresses[], unsigned int num_addresses)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_spas: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_spas(handle, addresses, num_addresses);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	ptr = SWIG_NewPointerObj((void *)addresses, SWIGTYPE_p_globus_ftp_control_host_port_t, 0);
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1, ptr);
    }   
    
    return tuple;  



}
PyObject *
local_spor(globus_ftp_control_handle_t * handle, globus_ftp_control_host_port_t addresses[], unsigned int num_addresses)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_spor: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_spor(handle, addresses, num_addresses);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;  

}

PyObject *
local_mode(globus_ftp_control_handle_t * handle, char  mode)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    globus_ftp_control_mode_t md;

    if (mode == '0')
    {
	md = GLOBUS_FTP_CONTROL_MODE_NONE;
    }
    else if (mode == 'S')
    {
	md = GLOBUS_FTP_CONTROL_MODE_STREAM;
    }    
    else if (mode == 'B')
    {
	md = GLOBUS_FTP_CONTROL_MODE_BLOCK;
    }    
    else if (mode == 'E')
    {
	md = GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK;
    }
    else if (mode == 'C')
    {
	md = GLOBUS_FTP_CONTROL_MODE_COMPRESSED;
    }
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_mode: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_mode(handle, md);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;  

}

PyObject *
local_tcp_buffer(globus_ftp_control_handle_t * handle,globus_ftp_control_tcpbuffer_t *tcp_buffer)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_tcp_buffer: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_tcp_buffer(handle, tcp_buffer);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;  


}

PyObject *
local_dcau(globus_ftp_control_handle_t * handle,const globus_ftp_control_dcau_t *dcau,gss_cred_id_t *delegated_credential_handle)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_dcau: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_dcau(handle, dcau,*delegated_credential_handle );
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;  

}

PyObject *
local_pbsz(globus_ftp_control_handle_t * handle,unsigned long bufsize)
{

    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_pbsz PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_pbsz(handle,bufsize);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple; 
}
PyObject *
local_type(globus_ftp_control_handle_t * handle, char data_type)
{
    globus_result_t ret;
    globus_ftp_control_type_t type;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_type: PyTuple_New()");
	return NULL;
    }
    if (data_type == 'I')
    {
	type = GLOBUS_FTP_CONTROL_TYPE_IMAGE;
    }
    else if (data_type == 'A')
    {
	type = GLOBUS_FTP_CONTROL_TYPE_ASCII;
    }
    else
    {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", -1));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", "That type is not implemented only ASCII and IMAGE are implemented\n"));
	return tuple;
    }
    
    ret = globus_ftp_control_local_type(handle,type, 0 );
    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", ret));
    
    if(ret != GLOBUS_SUCCESS){
	
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;  

}


PyObject *
get_pbsz(globus_ftp_control_handle_t * handle)
{

    globus_result_t ret;
    PyObject *tuple, *pyBufSize;
    char * tmpstr;
    unsigned long * bufsize;
    bufsize = (unsigned long *)globus_malloc(sizeof(unsigned long));
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "get_pbsz PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_get_pbsz(handle,bufsize);

    if(ret != GLOBUS_SUCCESS)
    {
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	pyBufSize = PyLong_FromUnsignedLong(*bufsize);
	PyTuple_SET_ITEM(tuple, 1, pyBufSize);
    }   
    
    return tuple; 
}

PyObject *
local_stru(globus_ftp_control_handle_t * handle, globus_ftp_control_structure_t *structure)
{

    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;

    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "local_stru PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_local_stru(handle,*structure);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple; 
}
PyObject *
data_read(globus_ftp_control_handle_t * handle, PyObject * pyBuffer, globus_byte_t * buffer, int max_length, PyObject * pyfunc, PyObject * arg)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char* tmpstr;
    callbackStruct4* cbStruct;
     if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "data_read: PyTuple_New()");
	return NULL;
    }
     cbStruct = (callbackStruct4*)globus_malloc(sizeof(callbackStruct4));
     Py_XINCREF(pyfunc);
     Py_XINCREF(arg);
     Py_XINCREF(pyBuffer);
     cbStruct->callback = pyfunc;
     cbStruct->arg = arg;
     cbStruct->buffer = pyBuffer;
      Py_BEGIN_ALLOW_THREADS
	 ret = globus_ftp_control_data_read(handle,buffer, max_length, PythonCompleteDataCallback, cbStruct);
     Py_END_ALLOW_THREADS
	if(ret != GLOBUS_SUCCESS){
	    PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	    tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	    PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	    
	}
	else {
	    ptr = SWIG_NewPointerObj((void *)cbStruct, SWIGTYPE_p_callbackStruct4, 0);
	    PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	    PyTuple_SET_ITEM(tuple, 1, ptr);
	}   
	     
	return tuple;
}

PyObject *
command_init(char * raw_command, globus_ftp_control_auth_info_t * auth_info)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    globus_ftp_control_command_t * command;
    command = (globus_ftp_control_command_t *) globus_malloc(sizeof(globus_ftp_control_command_t));

    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "command_init: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_command_init(command, raw_command, auth_info);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple; 
}

PyObject *
command_destroy(globus_ftp_control_command_t * command)
{
    globus_result_t ret;
    PyObject *tuple;
    char * tmpstr;
    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "command_destroy: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_command_destroy(command);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,Py_BuildValue("s", "NONE"));
    }   
    
    return tuple;
}

PyObject *
command_copy(globus_ftp_control_command_t * src)
{
    globus_result_t ret;
    PyObject *tuple, *ptr;
    char * tmpstr;
    globus_ftp_control_command_t * dest;
    dest = (globus_ftp_control_command_t *)globus_malloc(sizeof(globus_ftp_control_command_t));

    if (!(tuple=PyTuple_New(2)))
    {
	PyErr_SetString(PyExc_RuntimeError, "command_copy: PyTuple_New()");
	return NULL;
    }
    ret = globus_ftp_control_command_copy(dest, src);
    if(ret != GLOBUS_SUCCESS){
	PyTuple_SET_ITEM(tuple, 0,Py_BuildValue("i", -1));
	tmpstr = globus_object_printable_to_string(globus_error_get(ret));
	PyTuple_SET_ITEM(tuple, 1, Py_BuildValue("s", tmpstr));
        globus_free(tmpstr);
	
    }
    else {
	ptr =  SWIG_NewPointerObj((void *)dest, SWIGTYPE_p_globus_ftp_control_command_t, 0);
	PyTuple_SET_ITEM(tuple, 0, Py_BuildValue("i", 0));
	PyTuple_SET_ITEM(tuple, 1,ptr);
    }   
    
    return tuple;
}

PyObject *
eof_array_init(int size)
{
    int * eof_array;
    PyObject * ptr;
    eof_array = (int *)globus_malloc(sizeof(int) * size);
    ptr =  SWIG_NewPointerObj((void *)eof_array, SWIGTYPE_p_int, 0);
    return ptr;

}

void
eof_array_destroy(int * eof_array)
{
    globus_libc_free(eof_array);
}

PyObject * 
host_port_init(PyObject *ipAddr, unsigned short port)
{
    PyObject * ptr;
    int i = 0;
    int listOK = 0;
    int intOK = 0;
    int tmpInt;
    
    PyObject * pyInt;
    globus_ftp_control_host_port_t * host_port;
    host_port = (globus_ftp_control_host_port_t *)globus_malloc(sizeof(globus_ftp_control_host_port_t));
    

    listOK = PyList_Check(ipAddr);
    if (!listOK)
    {
	PyErr_SetString(PyExc_RuntimeError, "first argument must be a list object");
	return NULL;
    }
    
    for (i = 0; i < 4; i++)
    {
	pyInt  = PyList_GET_ITEM(ipAddr, i);
	intOK = PyArg_Parse(pyInt, "i", &tmpInt);
	if (!intOK)
	{
	    PyErr_SetString(PyExc_RuntimeError, "elements of list must be ints");
	    return NULL;
	}
	
	host_port->host[i] = tmpInt;
    }
    host_port->port = port;
    ptr =  SWIG_NewPointerObj((void *)host_port, SWIGTYPE_p_globus_ftp_control_host_port_t, 0);
    return ptr;
}

void
host_port_destroy(globus_ftp_control_host_port_t * host_port)
{
    globus_libc_free(host_port);
} 


PyObject *
host_port_info(globus_ftp_control_host_port_t * host_port)
{
    PyObject * tuple;
    int i;
    
    if (!(tuple=PyTuple_New(5)))
    {
	PyErr_SetString(PyExc_RuntimeError, "host_port_info: PyTuple_New()");
	return NULL;
    }
    for(i = 0; i <4; i++)
    {
	PyTuple_SET_ITEM(tuple, i, Py_BuildValue("i", host_port->host[i]));
    }
    PyTuple_SET_ITEM(tuple, 4, Py_BuildValue("i", host_port->port));
    return tuple;
    
}

PyObject *
host_port_array_init(int indeces)
{
    globus_ftp_control_host_port_t * hp_array;
    PyObject * ptr;
    hp_array = (globus_ftp_control_host_port_t *)globus_malloc(sizeof(globus_ftp_control_host_port_t) * indeces);
    ptr = SWIG_NewPointerObj((void *)hp_array, SWIGTYPE_p_globus_ftp_control_host_port_t, 0);
    return ptr;
}

void
host_port_array_destroy(globus_ftp_control_host_port_t * hp_array)
{ 
   globus_libc_free(hp_array);
}



void
examine_response(globus_ftp_control_response_t * response)
{
    printf("in examine_response");
    printf("%s\n", response->response_buffer);
}


%}
