#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, time, urlparse, optparse
import ZSI
from ConfigParser import ConfigParser
import pyGridWare.generated.stubs.Counter_services as COUNTER


###########################################################################
# GLOBALS
###########################################################################
TEST = 'CounterService'
CONFIG_FILE = 'config.txt'
SEC_PATHS = 'Paths'
SEC_LOCATION = 'Location'

CP = ConfigParser()
CP.read(CONFIG_FILE)
HOST = CP.get(SEC_LOCATION, 'host')
PORT = CP.get(SEC_LOCATION, 'port')


###########################################################################
# OPTIONS
###########################################################################
op = optparse.OptionParser()
op.add_option("-s", "--host",
              action="store", dest="host", default=HOST,
              help="location of service container")
op.add_option("-p", "--port",
              action="store", dest="port", default=PORT,
              help="port of service container")
(OPTIONS, ARGS) = op.parse_args()


############################################################################
# Counter
#     add
#     createCounter
############################################################################
def create(port):
    request = COUNTER.CreateCounterRequest()
    msg = port.createCounter(request)
    print "----- create ----- "
    print "\tAddress -- ", msg._EndpointReference._Address
    print "\tRefProp -- ", msg._EndpointReference._ReferenceProperties
    print "\t\tany -- ", msg._EndpointReference._ReferenceProperties._any
    return msg

def add(port, value=999):
    print "----- add ----- "
    request = value
    msg = port.add(request)
    print "RESPONSE: ", msg


############################################################################
# ResourceLifetime
#     Destroy
#     GetCurrentMessage
#     SetTerminationTime
############################################################################
def Destroy(port):
    print "----- destroy ----- "
    request = COUNTER.DestroyRequest()
    msg = port.Destroy(request)
    print "RESPONSE: ", msg

def SetTerminationTime(port, interval):
    print "----- SetTerminationTime ----- "
    request = COUNTER.SetTerminationTimeRequest()
    request._RequestedTerminationTime = time.gmtime(time.time()+interval)
    msg = port.SetTerminationTime(request)
    print "TerminationTime: ", msg._NewTerminationTime
    print "CurrrentTime:    ", msg._CurrentTime


############################################################################
# ResourceProperties
#     GetResourceProperty
############################################################################
def GetResourceProperty(port):
    print "----- GetResourceProperty ----- "
    class QualifiedName(tuple):
        '''Container class for immutable tuple representing a QName.
        '''
    request = QualifiedName(("http://counter.com", "Value"))
    msg = port.GetResourceProperty(request)
    print "RESPONSE: ", msg._any



def RunCounter(url):
    path = urlparse.urlparse(url)[2]
    kw = {'tracefile':sys.stdout, 'url':path}

    scheme,netloc,path,params,query,fragment = urlparse.urlparse(url)
    locator = COUNTER.CounterServiceLocator()
    port = locator.getCounterPortType(portAddress=url, **kw)

    response = create(port)
    port.endPointReference = response._EndpointReference

    add(port, 11)
    GetResourceProperty(port)
    add(port, 13)
    Destroy(port)
    try:
        add(port, 11)
    except ZSI.FaultException, ex:
        print "TERMINATED: Operation Failed after resource was destroyed."
    else:
        print "RESOURCE FAILED TO TERMINATE"
    print '*'*50

 
def main():
    netloc = '%s:%s'%(OPTIONS.host,OPTIONS.port)
    path = CP.get(SEC_PATHS, 'CounterService')
    urlTuple = ('http',netloc,path,'','','')
    url = urlparse.urlunparse(urlTuple)
    RunCounter(url)


if __name__ == '__main__':
    main()

