#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, time, urlparse, optparse
from twisted.internet import reactor
from ConfigParser import ConfigParser
import pyGridWare.generated.stubs.Counter_services as COUNTER
from pyGridWare.notification.NotificationConsumer import NotificationConsumer
from twisted.python import threadable
threadable.init()


###########################################################################
# GLOBALS
###########################################################################
DIALECT = "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
TEST = 'CounterService'
CONFIG_FILE = 'config.txt'
SEC_PATHS = 'Paths'
SEC_LOCATION = 'Location'

CP = ConfigParser()
CP.read(CONFIG_FILE)
HOST = CP.get(SEC_LOCATION, 'host')
PORT = CP.get(SEC_LOCATION, 'port')


###########################################################################
# OPTIONS
###########################################################################
op = optparse.OptionParser()
op.add_option("-s", "--host",
              action="store", dest="host", default=HOST,
              help="location of service container")
op.add_option("-p", "--port",
              action="store", dest="port", default=PORT,
              help="port of service container")
(OPTIONS, ARGS) = op.parse_args()



class QualifiedName(tuple):
    '''Use to package data for QName typecode, this
    way the data is not recognized as a sequence type.
    '''
    def __init__(self, value):
        tuple.__init__(self,value)


class TopicExpression:
    '''
    '''
    def __init__(self):
        self._attrs = {}
        self._attrs['Dialect'] = DIALECT

    def setMixed(self, value):
        self._text = value


"""
Counter
    add
    createCounter
"""
def create(port):
    request = COUNTER.CreateCounterRequest()
    msg = port.createCounter(request)
    print "----- create ----- "
    print "\tAddress -- ", msg._EndpointReference._Address
    print "\tRefProp -- ", msg._EndpointReference._ReferenceProperties
    print "\t\tany -- ", msg._EndpointReference._ReferenceProperties._any
    return msg

def add(port, value=999):
    print "----- add ----- "
    request = value
    msg = port.add(request)
    print "RESPONSE: ", msg

"""
ResourceLifetime
    Destroy
    GetCurrentMessage
    SetTerminationTime
"""
def Destroy(port):
    print "----- destroy ----- "
    request = COUNTER.DestroyRequest()
    msg = port.Destroy(request)
    print "RESPONSE: ", msg

def SetTerminationTime(port, interval):
    print "----- SetTerminationTime ----- "
    request = COUNTER.SetTerminationTimeRequest()
    request._RequestedTerminationTime = time.gmtime(time.time()+interval)
    msg = port.SetTerminationTime(request)
    print "TerminationTime: ", msg._NewTerminationTime
    print "CurrrentTime:    ", msg._CurrentTime


"""
ResourceProperties
    GetResourceProperty
"""
def GetResourceProperty(port):
    print "----- GetResourceProperty ----- "
    request = QualifiedName(("http://counter.com", "Value"))
    msg = port.GetResourceProperty(request)
    print "RESPONSE: ", msg._any


def Subscribe(port, consumer):
    """
    port -- Subscription Client
    consumer -- Notification Consumer
    Need to specify the prefix-namespace mapping in the
    XML instance to deal with QNames passed via the 
    TopicExpression mixed text content.
    """
    print "----- Subscribe ----- "
    
    request = COUNTER.SubscribeRequest()
    epr = consumer.context.getConsumerReference()
    request._ConsumerReference = epr
    request._TopicExpression = TopicExpression()
    request._TopicExpression.setMixed("counter:Value")
    request._UseNotify = True
    #request._Precondition = None
    #request._Selector = None
    #request._SubscriptionPolicy = None
    #request._InitialTerminationTime = None
    port.binding.nsdict['counter'] = "http://counter.com"
    msg = port.Subscribe(request)
    print "RESPONSE: ", msg



def counter(url):
    path = urlparse.urlparse(url)[2]
    kw = {'tracefile':sys.stdout, 'url':path}

    scheme,netloc,path,params,query,fragment = urlparse.urlparse(url)
    locator = COUNTER.CounterServiceLocator()

    port = locator.getCounterPortType(portAddress=url, **kw)
    response = create(port)

    # Get New port w/ EPR
    port = locator.getCounterPortType(portAddress=url, \
        endPointReference=response._EndpointReference, \
        **kw)

    consumer = NotificationConsumer()
    consumer.start()

    print "=======> Queue Events.."
    reactor.callWhenRunning(Subscribe, port, consumer)
    reactor.callWhenRunning(add, port, 13)
    reactor.callLater(5, reactor.stop)

 
def main():
    netloc = '%s:%s'%(OPTIONS.host, OPTIONS.port)
    path = CP.get(SEC_PATHS, 'CounterService')
    urlTuple = ('http',netloc,path,'','','')
    url = urlparse.urlunparse(urlTuple)
    counter(url)


if __name__ == '__main__':
    reactor.callWhenRunning(main)
    reactor.run()



