#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, sha, time, urlparse, optparse
from ConfigParser import ConfigParser

import ZSI.TC
from ZSI.wstools.Namespaces import OASIS
from pyGridWare.generated.types import GetNSContainer
import pyGridWare.generated.stubs.Counter_services as COUNTER_SERVICE
import pyGridWare.generated.types.Counter_services_types as COUNTER_TYPES
from pyGridWare.security.gss.GssSignatureHandler import GssSignatureHandler
from pyGridWare.security.gss.Authenticate import GssAuthenticationHandler
from pyGridWare.security.Utility import MessageSigUtility


###########################################################################
# GLOBALS
###########################################################################
CONFIG_FILE = 'config.txt'
CP = ConfigParser()
CP.read(CONFIG_FILE)
SEC_PATHS = 'Paths'
SEC_LOCATION = 'Location'
HOST = CP.get(SEC_LOCATION, 'host')
PORT = CP.get(SEC_LOCATION, 'port')


###########################################################################
# OPTIONS
###########################################################################
op = optparse.OptionParser()
op.add_option("-s", "--host",
              action="store", dest="host", default=HOST,
              help="location of service container")
op.add_option("-p", "--port",
              action="store", dest="port", default=PORT,
              help="port of service container")

(options, args) = op.parse_args()


###########################################################################
# Counter
#    create
#    add
###########################################################################
def create(port):
    print "----- create ----- "
    request = COUNTER_SERVICE.CreateCounterRequest()
    MessageSigUtility.Sign(request)
    msg = port.createCounter(request)
    return msg

def add(port, value=999):
    '''Since request is an immutable, must grab the typecode for
    request to be marked as signed.
    '''
    print "----- add ----- "
    class SignInt(int):
        '''Container class for immutable int represent a xsd:int.
        '''
    request = SignInt(value)
    MessageSigUtility.Sign(request, COUNTER_SERVICE.AddInputMessage)
    msg = port.add(request)
    print "RESPONSE: ", msg


###########################################################################
# ResourceLifetime
#     Destroy
#     GetCurrentMessage
#     SetTerminationTime
###########################################################################
def Destroy(port):
    print "----- destroy ----- "
    request = COUNTER_SERVICE.DestroyRequest()
    MessageSigUtility.Sign(request)
    msg = port.Destroy(request)
    print "RESPONSE: ", msg

def SetTerminationTime(port, interval):
    print "----- SetTerminationTime ----- "
    request = COUNTER_SERVICE.SetTerminationTimeRequest()
    request._RequestedTerminationTime = time.gmtime(time.time()+interval)
    MessageSigUtility.Sign(request)
    msg = port.SetTerminationTime(request)
    print "RESPONSE: ", msg


###########################################################################
# ResourceProperties
#     GetResourceProperty
###########################################################################
def GetResourceProperty(port):
    print "----- GetResourceProperty ----- "
    class QualifiedName(tuple):
        '''Container class for immutable tuple representing a QName.
        '''
    request = QualifiedName(("http://counter.com", "Value"))
    MessageSigUtility.Sign(request, COUNTER_SERVICE.GetResourcePropertyRequest)
    msg = port.GetResourceProperty(request)
    print "RESPONSE: ", msg


def RunCounter(url, path, KeyName):
    kw = {'tracefile':sys.stdout, 'url':path}
    scheme,netloc,path,params,query,fragment = urlparse.urlparse(url)
    locator = COUNTER_SERVICE.CounterServiceLocator()

    sig_handler = GssSignatureHandler()
    sig_handler.setContextID(KeyName)
    kw['sig_handler'] = sig_handler

    port = locator.getCounterPortType(portAddress=url, **kw)
    response = create(port)
    epr = response._EndpointReference

    port = locator.getCounterPortType(portAddress=url, endPointReference=epr, **kw)
    add(port, 11)
    SetTerminationTime(port, 5)
    add(port, 13)
    GetResourceProperty(port)
    print '*'*50
    print 'Wait for resource to terminate, next operation should fail'
    time.sleep(15)
    try:
        Destroy(port)
    except ZSI.FaultException, ex:
        print "TERMINATED: Operation Failed after resource was destroyed."
    else:
        print "RESOURCE FAILED TO TERMINATE"
    print '*'*50

    
def main():
    path = CP.get(SEC_PATHS, 'AuthService')
    netloc = '%s:%s'%(options.host,options.port)
    urlTuple = ('http',netloc,path,'','','')
    url = urlparse.urlunparse(urlTuple)
    
    gssAuth = GssAuthenticationHandler()
    KeyName = gssAuth.authenticate(url, tracefile=sys.stdout)
    path = CP.get(SEC_PATHS, 'SecureCounterService')
    urlTuple = ('http',netloc,path,'','','')
    url = urlparse.urlunparse(urlTuple)
    RunCounter(url, path, KeyName)


if __name__ == '__main__':
    main()

