#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os, sys, socket
from ConfigParser import ConfigParser
from ZSI.ServiceContainer import ServiceContainer
from twisted.internet import reactor
from twisted.python import log
from twisted.python import threadable
threadable.init()
from ZSI.wstools.logging import ILogger, setLoggerClass, setBasicLogger

os.environ["USE_TWISTED"] = "0"
CONFIG_FILE = 'config.txt'
SEC_PATHS = 'Paths'
SEC_CLASS = 'ClassName'
SEC_LOCATION = 'Location'
SEC_CONTAINER = 'Container'
SEC_LOGGING = 'logging'

#from logging import Logger, setLoggerClass
class TwistedLogger(ILogger):
    def __init__(self, name):
        self.name = name
    def warning(self, *args):
        msg = args[0]
        if len(args) > 1: msg = args[0] %args[1:]
        log.msg(self.WARN %self.name, msg)
    WARN  = 'WARNING %s:  '
    def debug(self, *args):
        msg = args[0]
        if len(args) > 1: msg = args[0] %args[1:]
        log.debug(self.DEBUG %self.name, msg)
    DEBUG = 'DEBUG %s:  '
    def error(self, *args):
        msg = args[0]
        if len(args) > 1: msg = args[0] %args[1:]
        log.error(self.ERROR %self.name, msg)
    ERROR = 'ERROR %s:  '
#setLoggerClass(TwistedLogger)
setBasicLogger()


class TwistedMixIn:
    '''Mix-in class to handle each request in a thread.'''

    def process_request_thread(self, request, client_address):
        '''Same as in BaseServer but as a thread.
        In addition, exception handling is done here.
        '''
        try:
            self.finish_request(request, client_address)
            self.close_request(request)
        except:
            self.handle_error(request, client_address)
            self.close_request(request)

    def process_request(self, request, client_address):
        '''Start a new thread to process the request.
        '''
        target = self.process_request_thread
        reactor.callInThread(target, request, client_address)


class ServiceContainerTwisted(TwistedMixIn, ServiceContainer):

    def setShutdown(self, on):
        assert on in (True,False)
        self._shutdown = on
    def getShutdown(self):
        return self._shutdown
    shutdown = property(getShutdown, setShutdown, None, "shutting down")
    
    def serve_forever(self):
        '''Handle one request at a time until doomsday.
        '''
        self.shutdown = False
        while not self.shutdown:
            log.msg("==> handle_request")
            self.handle_request()

    def server_close(self):
        '''Called to clean-up the server.
        '''
        self.shutdown = True
        self.socket.shutdown(2)


def StartServer(port=80, services=()):
    '''port --
       services -- list of service instances
    '''
    address = ('', port)
    sc = ServiceContainerTwisted(address)
    for service in services:
        path = service.getPost()
        sc.setNode(service, path)

    reactor.addSystemEventTrigger('during', 'shutdown', sc.server_close)
    reactor.callInThread(sc.serve_forever)
    reactor.run()


def main():
    services = []
    cp = ConfigParser()
    cp.read(CONFIG_FILE)
    for key in cp.get(SEC_CONTAINER, "services").split():
        path = cp.get(SEC_PATHS, key)
        module,className = eval(cp.get(SEC_CLASS, key))
        exec("from %s import %s as SERVICE_CLASS" %(module, className))
        services.append(SERVICE_CLASS(path))

    if cp.getint(SEC_LOGGING, 'debug') == 1:
        log.startLogging(sys.stdout)

    StartServer(cp.getint(SEC_LOCATION,'port'), services)


if __name__ == '__main__':
    main()
