############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import socket
from ConfigParser import ConfigParser
from twisted.application import service, internet
from twisted.protocols import wire
from twisted.internet import protocol
from twisted.web.server import Site
from twisted.web.resource import Resource

from pyGridWare.services.CounterService import Counter
from pyGridWare.services.ManagedJobService import ManagedJobService
from pyGridWare.services.ManagedJobFactoryService import ManagedJobFactoryService
from pyGridWare.services.SecureCounterService import SecureCounter
from pyGridWare.services.SecureConversationService import SecureConversationService
from pyGridWare.services.SubscriptionManager import SubscriptionManager
from pyGridWare.services.WidgetService import Widget


###########################################################################
# Grab Port from configuration file
#
###########################################################################
CP = ConfigParser()
CP.read('config.txt')
PORT = CP.getint('Location', 'port')
HOST = socket.gethostbyname(socket.gethostname())


###########################################################################
# Get Resources
#
###########################################################################
def GetResource():
    root = Resource()
    root.putChild('wsrf', Resource())
    root.putChild('widget', Widget('/widget'))

    resource = Resource()
    root.getStaticEntity('wsrf').putChild('services', resource)

    resource.putChild('CounterService', 
        Counter(post='/wsrf/services/CounterService'))

    resource.putChild('SecureCounterService-authService', 
        SecureConversationService(post='/wsrf/services/SecureCounterService-authService'))

    resource.putChild('SecureCounterService', 
        SecureCounter(post='/wsrf/services/SecureCounterService'))

    # ManagedJobService is not WORKING
    resource.putChild('ManagedJobService', 
        ManagedJobService('/wsrf/services/ManagedJobService'))

    # ManagedJobFactoryService is not WORKING
    resource.putChild('ManagedJobFactoryService', 
        ManagedJobFactoryService('/wsrf/services/ManagedJobFactoryService'))


    # XXX Need to Set URL for SubscriptionManager, need to know
    # its URL before it has received a request.
    #subResource = SubscriptionManagerService(post='/wsrf/services/SubscriptionManagerService')
    #subResource.setServiceURL('http://%s:%d%s' %(port.getHost().host, port.getHost().port, sub_path)
    sub_path='/wsrf/services/SubscriptionManagerService'
    subResource = SubscriptionManager.getInstance(post=sub_path)
    subResource.setServiceURL('http://%s:%d/wsrf/services/SubscriptionManagerService' %(HOST,PORT))
    resource.putChild('SubscriptionManagerService', subResource)
    return root


###########################################################################
# Set up Application
#
###########################################################################
myFactory = Site(GetResource())
port = internet.TCPServer(PORT, myFactory)
application = service.Application("WSRF-Container")
port.setServiceParent(application)


