############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys
from twisted.internet import reactor
from twisted.python import log
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.generated.stubs.WS_BaseNotification_services import \
        Notify, NotificationConsumerSOAPBindingSOAP, WS_BaseNotificationServiceLocator
from pyGridWare.generated.types.NotificationConsumer_services_types import ns7
#from pyGridWare.generated.types.WS_BaseNotification_services_types import \
#    docs_oasis_open_org_wsrf_2004_06_wsrf_WS_ResourceProperties_1_2_draft_01_xsd as ns7

class NotificationMessage:
    '''
  <Notify xmlns="http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd">
    <NotificationMessage>
      <Topic Dialect="http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
              xmlns:ns1="http://counter.com">ns1:Value</Topic>
      <ProducerReference>
        <wsa:Address>http://192.168.1.101:8080/wsrf/services/CounterService</wsa:Address>
        <wsa:ReferenceProperties>
          <ns2:CounterKey xmlns:ns2="http://counter.com">4163038</ns2:CounterKey>
        </wsa:ReferenceProperties>
      </ProducerReference>
      <Message xsi:type="ns3:ResourcePropertyValueChangeNotificationElementType"
                xmlns:ns3="http://wsrf.globus.org/core/notification">
        <ns4:ResourcePropertyValueChangeNotification
              xmlns:ns4="http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.xsd">
          <ns4:NewValue>
            <ns5:Value xmlns:ns5="http://counter.com">3</ns5:Value>
          </ns4:NewValue>
        </ns4:ResourcePropertyValueChangeNotification>
      </Message>
    </NotificationMessage>
  </Notify>
    '''
    class Topic:
        def __init__(self):
            ''' XXX ns2:Value is BOGUS!!!  Create a typecode "SimpleTopicExpression" that will parse
            and serialize the mixed text content as a QName, it is a xsd:token according to the
            WS-Topics specification.
            '''
            self._attrs = {'Dialect':'http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple'}
            self._text = 'ns2:Value'

    class Message:
        typecode = ns7.ResourcePropertyValueChangeNotificationElementType_Def(pname='Message')
        class ResourcePropertyValueChangeNotification:
            class NewValue: pass
            def __init__(self):
                self._NewValue = self.__class__.NewValue()
        def __init__(self):
            self._ResourcePropertyValueChangeNotification = self.__class__.ResourcePropertyValueChangeNotification()

    class ProducerReference: pass
    def __init__(self):
        self._Topic = NotificationMessage.Topic()
        self._ProducerReference = self.__class__.ProducerReference()
        self._Message = self.__class__.Message()



class NotificationProducer:
    '''The NotificationProducer can act as a SubscriptionManager, 

    request -- subscription request, containing ConsumerReference and TopicExpression.
    ctx -- resource context subscription applies to.
<Subscribe xmlns="http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd">
  <ConsumerReference>
    <wsa:Address>http://192.168.1.101:52964/wsrf/services/NotificationConsumerService</wsa:Address>
    <wsa:ReferenceProperties>
       <ns1:NotificationConsumerKey xmlns:ns1="http://www.globus.org/namespaces/2004/06/core">522238</ns1:NotificationConsumerKey>
    </wsa:ReferenceProperties>
  </ConsumerReference>
  <TopicExpression Dialect="http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple" xmlns:ns2="http://counter.com">ns2:Value</TopicExpression>
  <UseNotify>true</UseNotify></Subscribe>
    '''
    subscriptions = []

    def __init__(self):
        '''
        stub -- NotificationConsumer RPC instance
        subscription -- Subscription instance
        '''
        self.stub = None

    def setUp(self, subscription):
        '''
        request -- ConsumerReference element declaration, contents
            ConsumerReference, TopicExpression, UseNotify
        '''
        request = subscription.request
        #if subscription.request._UseNotify is True:
        #    reference = subscription.reference
        #    url = reference._Address

        #   topic = subscription.request._TopicExpression
        #   subscription.producer.properties.addTopicListener(subscription.Notify, subscription.topic)

        url = subscription.reference._Address
        locator = WS_BaseNotificationServiceLocator()
        self.stub = locator.getNotificationConsumer(url, \
            endPointReference=request._ConsumerReference, tracefile=sys.stdout)

        # REMOVE 
        self.subscription = subscription
        subscription.producer.properties.addTopicListener(self.Notify, subscription.topic)


    def Notify(self, property):
        '''Send Notify message.

  <Notify xmlns="http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd">
    <NotificationMessage>
      <Topic Dialect="http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
              xmlns:ns1="http://counter.com">ns1:Value</Topic>
      <ProducerReference>
        <wsa:Address>http://192.168.1.101:8080/wsrf/services/CounterService</wsa:Address>
        <wsa:ReferenceProperties>
          <ns2:CounterKey xmlns:ns2="http://counter.com">4163038</ns2:CounterKey>
        </wsa:ReferenceProperties>
      </ProducerReference>
      <Message xsi:type="ns3:ResourcePropertyValueChangeNotificationElementType"
                xmlns:ns3="http://wsrf.globus.org/core/notification">
        <ns4:ResourcePropertyValueChangeNotification
              xmlns:ns4="http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceProperties-1.2-draft-01.xsd">
          <ns4:NewValue>
            <ns5:Value xmlns:ns5="http://counter.com">3</ns5:Value>
          </ns4:NewValue>
        </ns4:ResourcePropertyValueChangeNotification>
      </Message>
    </NotificationMessage>
  </Notify>
        '''
        log.msg("Notify ---> ")
        request = Notify()
        request._NotificationMessage = NotificationMessage()
        request._NotificationMessage._ProducerReference = \
            AddressingUtils.createEndpointReference(self.subscription.producer)
        request._NotificationMessage._Message._ResourcePropertyValueChangeNotification._NewValue._any = \
            property
        log.msg("<-- Notify ")
        self.stub.Notify(request)



