############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys
from twisted.python import log
from pyGridWare.patterns import Singleton
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.notification.NotificationProducer import NotificationProducer
from pyGridWare.resource.ResourceContext import ResourceContext
from pyGridWare.lifetime.ResourceLifetime import ImmediateResourceTermination,\
    ScheduledResourceTermination

#from pyGridWare.generated.types.BaseNotification_types import ns6
from pyGridWare.generated.types.WS_BaseNotification_services_types import *
from pyGridWare.generated.stubs.WS_BaseNotification_services import \
    WS_BaseNotificationServiceLocator, NotificationConsumerSOAPBindingSOAP
#from pyGridWare.generated.types.NotificationConsumer_services_types import ns7

from pyGridWare.generated.resource.SubscriptionProducer_context import \
    SubscriptionContext


"""
class NotificationMessage:
    class Topic:
        def __init__(self):
            ''' XXX ns2:Value is BOGUS!!! 
            '''
            self._attrs = {'Dialect':'http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple'}
            self._text = 'ns2:Value'

    class Message:
	typecode = ns7.ResourcePropertyValueChangeNotificationElementType_Def(pname='Message')
        class ResourcePropertyValueChangeNotification:
            class NewValue: pass
            def __init__(self):
                self._NewValue = self.__class__.NewValue()
        def __init__(self):
            self._ResourcePropertyValueChangeNotification = self.__class__.ResourcePropertyValueChangeNotification()

    class ProducerReference: pass
    def __init__(self):
        self._Topic = NotificationMessage.Topic() 
        self._ProducerReference = self.__class__.ProducerReference()
	self._Message = self.__class__.Message()
"""

 

class Topic:
    '''A Topic must specify a "Dialect", which is used to
    interpret the "expression" passed via mixed or <any>.
    '''
    simple = "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
    def __init__(self, topic):
        '''
        element wildcard
        mixed content
        attributes: Dialect
        '''
        self.topic = topic

    def getDialect(self):
        return self.topic._attrs['Dialect']

    def getRootTopic(self):
        return self.topic._text

    def evaluate(self, event):
        '''XXX Need DOM to resolve prefix
        The only valid token for 'Simple' dialect is a qName, but
        sadly this is not represented as a qName in XMLSchema.  Instead
        it is only understood to be a qName in this layer.
        '''
        if self.simple != self.getDialect():
            raise NotImplementedError, 'Dialect %s currently not supported' %self.dialect
        qualifiedName = self.getRootTopic()
   

class Subscription:
    '''Represents a subscription, registered to NotificationProducer
    '''
    def __init__(self, context, producer):
        '''
        context -- SubscriptionContext instance.
        producer -- Monitoring Resource Context 
        '''
        self.context = context
        self.producer = producer
        self.request = None
        self.topic = None

    def getContext(self):
        return self.__sub_context
    def setContext(self, context):
        assert isinstance(ctx, SubscriptionContext), \
            'resource must be a SubscriptionContext instance'
        self.__sub_context = context
    context = property(getContext, setContext, None,
        'Subscription Context')
 
    def getProducerContext(self):
        return self.__p_context
    def setProducerContext(self, context):
        assert isinstance(ctx, ResourceContext), \
            'resource must be a ResourceContext instance'
        self.__p_context = context
    producer = property(getProducerContext, setProducerContext, None,
        'Subscribed to Resource Context')

    def getSubscriptionRequest(self):
        return self.__context
    def setSubscriptionRequest(self, request):
        assert request is None or isinstance(request, ns6.Subscribe_Dec), \
            'request must be a Subscribe declaration (%s,%s)' \
            %(ns6.Subscribe_Dec.schema,ns6.Subscribe_Dec.literal)
        self.__request = request
    request = property(getSubscriptionRequest, setSubscriptionRequest, None,
        'Subscription Request')

    #def getNotificationConsumerStub(self):
    #    return self.__stub
    #def setNotificationConsumerStub(self, stub):
    #    assert ctx is None or isinstance(stub, NotificationConsumerSOAPBindingSOAP), \
    #        'stub must be a NotificationConsumerSOAPBindingSOAP instance'
    #    self.__stub = stub
    #stub = property(getNotificationConsumerStub, setNotificationConsumerStub, None,
    #    'NotificationConsumer stub')

    def getTopicExpression(self):
        return self.__topic
    def setTopicExpression(self, topic):
        assert topicExpression is None or isinstance(topicExpression, ns6.TopicExpression_Dec), \
            'topicExpression must be a TopicExpression instance'
        assert topicExpression is None or self.__topic._attrs.has_key('Dialect'), \
            'topicExpression must have Dialect attribute'
        self.__topic = topic
    topic = property(getTopicExpression, setTopicExpression, None,
        'TopicExpression')

    def getConsumerReference(self):
        return self.__reference
    def setConsumerReference(self, reference):
        assert reference is None or isinstance(reference, ns6.ConsumerReference_Dec), \
            'reference must be a ConsumerReference'
        self.__reference = reference
    reference = property(getConsumerReference, setConsumerReference, None,
            'EPR (%s,%s)' %(ns6.ConsumerReference_Dec.schema,ns6.ConsumerReference_Dec.literal))

    def setUp(self, request):
        '''
        request -- ConsumerReference element declaration, contents
            ConsumerReference, TopicExpression, UseNotify
        '''
        self.request = request
        if self.request._UseNotify is True:
            self.reference = self.request._ConsumerReference
            url = self.reference._Address
            self.topic = self.request._TopicExpression
            #locator = WS_BaseNotificationServiceLocator()
            #self.stub = locator.getNotificationConsumer(url, \
            #    endPointReference=self.request._ConsumerReference, tracefile=sys.stdout)
            #self.producer.properties.addTopicListener(self.Notify, self.topic)
            self.notificationProducer = NotificationProducer()
            self.notificationProducer.setUp(self)

    #def Notify(self, property):
    #    '''Send Notify message.
    #    '''
    #    log.msg("Notify ---> ")
    #    request = Notify()
    #    request._NotificationMessage = NotificationMessage()
    #    request._NotificationMessage._ProducerReference = AddressingUtils.createEndpointReference(self.producer)
    #    request._NotificationMessage._Message._ResourcePropertyValueChangeNotification._NewValue._any = property
    #    log.msg("<-- Notify ")
    #    self.stub.Notify(request)


