#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, urlparse

import ZSI.TC
from ZSI.wstools.Namespaces import SOAP, BEA, WSSE,GLOBUS
from ZSI.utility import ElementProxy

from pyGridWare.generated.stubs.SecureConversation_services import *
from pyGridWare.generated.types.SecureConversation_services_types import *
#from pyGridWare.generated.types.Addressing_types import *
from pyGridWare.security.gss.GssSecureConversation import GssSecurityContext
from pyGridWare.security.gss.GssSignatureHandler import GssSignatureHandler
import base64

class RequestType(str):
        typecode = ns8.RequestType_Dec()
        
class Token(str):
        typecode = ns8.TokenType_Dec()

class BinaryExchange(str):
    typecode = ns8.BinaryExchange_Dec()
    def __new__(cls, value):
        s = str.__new__(cls, value)
        s._attrs = {}
        return s

class Holder:
    pass

class GssAuthenticationHandler:
    mechanism = "http://www.globus.org/2002/04/xmlenc#gssapi-gsi"
    namespaceURI = "http://wsrf.globus.org/security/authentication/secureconv/types"

    def getMechanism(cls):
        return cls.mechanism
    getMechanism = classmethod(getMechanism)

    def getNamespace(cls):
        return cls.namespaceURI
    getNamespace = classmethod(getNamespace)

    def continueCreateContext(self, port,ctx, response):
        ctxId = response._attrs['Context']
       
        serverTokens = response._any[(WSSE.TRUST,"BinaryExchange")][0]
        request = RequestSecurityTokenResponseMsg()
        request._any = []
        request._attrs = {}
        #I have to put the context Id into the ._attrs member
        request._attribute_typecode_dict = {}
        request._attribute_typecode_dict['Content'] = ZSI.TC.String()
        request._attrs['Context'] = ctxId
        tokens = ctx.initSecurityContext(serverTokens)
        binEx = BinaryExchange(tokens)
        request._any.append(binEx)
        request._any[-1]._attrs['EncodingType'] = \
                                             "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
        request._any[-1]._attrs['ValueType'] = \
                                          "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_GSI_TOKEN"
        response = port.RequestSecurityTokenResponse(request)
        return response

    def createContext(self,port,ctx):
        tokens = ctx.initSecurityContext()
        binEx = BinaryExchange(tokens)
        request = RequestSecurityTokenMsg()
        request._any = []
        request._any.append(RequestType('http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue'))
        request._any.append(Token('http://schemas.xmlsoap.org/ws/2004/04/security/sc/sct'))
        request._any.append(binEx)
        request._any[-1]._attrs['EncodingType'] = \
                                                "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
        request._any[-1]._attrs['ValueType'] = \
                                             "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_GSI_TOKEN"
        response = port.RequestSecurityToken(request)
        return response

    def authenticate(self, url, delegation=True, tracefile=None):
        context = GssSecurityContext()
        mechanism = GssAuthenticationHandler.getMechanism()
        namespaceURI = GssAuthenticationHandler.getNamespace()
        mechanism_attr_key = (namespaceURI,"mechanism-type")
        continue_attr_key = (namespaceURI,"continue-needed")

        if delegation is True:
            context.setDelegation()

        # Initialize Port instance
        locator = SecureConversationServiceLocator()
        scheme,netloc,path,params,query,fragment = urlparse.urlparse(url)
        kw = {'url':path}
        if tracefile is not None:
            kw['tracefile'] = tracefile
        port = locator.getSecureConversation(portAddress=url, **kw)

        
        response = self.createContext(port, context)
        #When the context has been created the response has
        #a SecurityContextToken
        done = False
        while not done:
            response = self.continueCreateContext(port,context,response)
            if response._any.has_key((BEA.SECCONV, 'SecurityContextToken')):
                done = True
                
        secureContextToken = response._any[(BEA.SECCONV, 'SecurityContextToken')][0]
        address = secureContextToken._any[(GLOBUS.SECCONV, 'ContextAddress')][0]
        referencePropertiesContent = address._ReferenceProperties._any
        contextID = referencePropertiesContent[(GLOBUS.CORE, 'SecureContextKey')][0]
        context.setContextID(contextID)
        home = GssSecurityContext.getResourceHome()
        home.create(context, contextID)
        return contextID


