############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
from pyGridWare.generated.services.SecureConversation_services_server import *
from pyGridWare.generated.types.SecureConversation_services_types import * 
from pyGridWare.security.gss.GssSecurityContext import GssSecurityContext
from ZSI.wstools.Namespaces import WSSE, BEA

#Move this somewhere else, the AuthenticationHandler and the SecureConversation
#need these classes
class Holder:
    pass

class RequestType(str):
        typecode = ns8.RequestType_Dec()

class Identifier:
    typecode =  schemas_xmlsoap_org_ws_2004_04_sc.Identifier_Dec()
    
class ContextAddress:
    typecode = wsrf_globus_org_core_2004_07_security_secconv.ContextAddress_Dec()

class SecurityContextToken:
    typecode = schemas_xmlsoap_org_ws_2004_04_sc.SecurityContextToken_Dec()

class SecureKey(str):
    typecode = ZSI.TC.String(pname=("http://www.globus.org/namespaces/2004/06/core","SecureContextKey "))

class Token(str):
        typecode = ns8.TokenType_Dec()

class BinaryExchange(str):
    typecode = ns8.BinaryExchange_Dec()
    def __new__(cls, value):
        s = str.__new__(cls, value)
        s._attrs = {}
        return s

class SecureConversationService(SecureConversation):
    
    def __init__(self, post='/wsrf/services/', **kw):
        SecureConversation.__init__(self, post)

    def wsa_RequestSecurityToken(self, ps, address):
        response = SecureConversation.wsa_RequestSecurityToken(self, ps, address)
        context = GssSecurityContext()
        tokens = self.request._any[(WSSE.TRUST,"BinaryExchange")][0]
        outTokens = context.acceptSecurityContext(tokens)
        contextID = context.hashCode()
        home = GssSecurityContext.getResourceHome()
        home.create(context, contextID)
        response._any = []
        response._attrs = {}
        response.attribute_typecode_dict['Context'] = ZSI.TC.String()
        response._attrs['Context'] = 'uuid:' + contextID
        response._any.append(RequestType('http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue'))
        response._any.append(Token('http://schemas.xmlsoap.org/ws/2004/04/security/sc/sct'))
        
        binEx = BinaryExchange(outTokens)
        response._any.append(binEx)
        response._any[-1]._attrs['EncodingType'] = \
                                                "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
        response._any[-1]._attrs['ValueType'] = \
                                             "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_GSI_TOKEN"                     
        return response

    def wsa_RequestSecurityTokenResponse(self, ps, address):
        response = SecureConversation.wsa_RequestSecurityTokenResponse(self, ps, address)
        tokens = self.request._any[(WSSE.TRUST,"BinaryExchange")][0]
        contextID = self.request._attrs['Context']
        home = GssSecurityContext.getResourceHome()
        context_id = contextID.split(':')[1]
        context = home.get(context_id)
        outTokens = context.acceptSecurityContext(tokens)
        response._attrs = {}
        response.attribute_typecode_dict['Context'] = ZSI.TC.String()
        response._attrs['Context'] =  contextID
        if context.contextCreated():
            response._any = SecurityContextToken()
            response._any._Identifier = contextID
            response._any._any = ContextAddress()
            response._any._any._Address = address._addressTo
            response._any._any._ReferenceProperties = Holder()
            response._any._any._ReferenceProperties._any = SecureKey(context_id)
        else:
            response._any = []
            binEx = BinaryExchange(outTokens)
            response._any.append(binEx)
            response._any[-1]._attrs['EncodingType'] = \
                                             "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
            response._any[-1]._attrs['ValueType'] = \
                                          "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_GSI_TOKEN"
        return response
