############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, string, urlparse
from distutils.core import setup
from distutils.spawn import spawn
from distutils.dist import Distribution
from os.path import isfile
from ConfigParser import ConfigParser
import ZSI
from StringIO import StringIO

GLOBUS_LOCATION = os.environ.get('GLOBUS_LOCATION')
ALLOW_SLOPPY_WSDL = True
CONFIG = 'config.txt'
GLOBUS_SECTION = 'GLOBUS_WSDL'
LOCAL_SECTION = 'LOCAL_WSDL'
DEBUG = False

def WriteServiceFile(wsdl):
    ss = ServerCallbackDescription()
    ss.fromWsdl(wsdl)
    file_name =  'pyGridWare/generated/services/%s.py' %ss.getStubName()
    fd = open(file_name, 'w+')
    ss.write(fd)
    fd.close()

class WritePyGridWareModule(ZSI.wsdl2python.WriteServiceModule):
    def get_module_names(self):
        f_types, f_services = ZSI.wsdl2python.WriteServiceModule.get_module_names(self)
        typesModule = 'pyGridWare/generated/types/%s' %f_types
        stubModule = 'pyGridWare/generated/stubs/%s' %f_services
        return typesModule, stubModule

    def write_services(self, f_types, f_services, fd, hasSchema):
        header = '%s \n# %s.py \n# generated by %s \n# \n# \n%s\n\n'\
                  %('#'*50, f_services, self.__module__, '#'*50)
        fd.write(header)
        moduleTypes = f_types.replace('/','.')
        schema_namespaces = []
        for schema in self.wsdl.types:
            schema_namespaces.append(schema.getTargetNamespace())

        for service in self._wa.getServicesList():
            sd = ZSI.wsdl2python.ServiceDescription(self._addressing)
            if hasSchema:
                sd.imports += ['\nfrom %s import *' % moduleTypes]
                sd.imports += ['\nfrom pyGridWare.utility.dom import DomletteReader, DomletteElementProxy']
                for ns in self.nsh.getNSList():
                    if ns not in schema_namespaces: continue
                    sd.imports += ['\nfrom %s import \\\n    %s as %s' \
                                  % ( moduleTypes,
                                      self.nsh.getModuleName(ns),
                                      self.nsh.getAlias(ns)) ]

            sd.fromWsdl(service, self.typeDict)
            if self._importlib:
                exec( 'import %s' % self._importlib )
                exec( 'obj = %s' % self._importlib )
                if hasattr(obj, 'clientimports' ):
                    sd.imports += obj.clientimports

            # XXX Quick FIX...
            # Insert writerClass and readerClass into __init__
            token = 'def __init__(self, addr, endPointReference=None, **kw):'
            stringPoints = sd.serviceBindings.split(token)
            id1 = ' '*8
            id2 = ' '*12
            tokIdx = 0
            s = StringIO()
            lastIdx = len(stringPoints) - 1
            while tokIdx < lastIdx:
                print "INDEX: ", tokIdx
                s.write(stringPoints[tokIdx])
                s.write(token)
                s.write('\n%sif not kw.has_key("readerclass"):\n%skw["readerclass"] = DomletteReader' %(id1,id2))
                s.write('\n%sif not kw.has_key("writerclass"):\n%skw["writerclass"] = DomletteElementProxy' %(id1,id2))
                tokIdx += 1
                s.write(stringPoints[tokIdx])

            sd.serviceBindings = s.getvalue()
            sd.write(fd)
        fd.write('\n')

def WriteStubsTypes(wsdl):
    wsm = WritePyGridWareModule(wsdl, addressing=True)
    wsm.write(False)


def GetWSDL(url, isGlobus=True):
    file = url
    if isGlobus:
        assert GLOBUS_LOCATION is not None, 'Must set GLOBUS_LOCATION to regenerate globus modules'
        location = GLOBUS_LOCATION + '/'
        file = urlparse.urljoin(location, url)
    reader = ZSI.wstools.WSDLTools.WSDLReader()
    if isfile(file) is False:
        return reader.loadFromURL(file)
    return reader.loadFromFile(file)


def RegenerateModules(services=False, stubs=False):
    '''Regenerate generated modules 
    '''
    if services is stubs is False:
        return
    cp = ConfigParser()
    cp.read(CONFIG)
    for s,f in cp.items(GLOBUS_SECTION):
        wsdl = GetWSDL(f)
        if services: WriteServiceFile(wsdl)
        if stubs: WriteStubsTypes(wsdl)
    for s,f in cp.items(LOCAL_SECTION):
        wsdl = GetWSDL(f, isGlobus=False)
        if services: WriteServiceFile(wsdl)
        if stubs: WriteStubsTypes(wsdl)


class ServerCallbackDescription:
    method_prefix = 'wsa'
    ID1 = '    '
    ID2 = ID1 * 2
    ID3 = ID1 * 3
    ID4 = ID1 * 4
    base_class = 'WSResource'
    base_class_mod = 'pyGridWare.wsrf.resource'
    stub_class_dir = 'pyGridWare.generated.stubs'

    def __init__(self):
        self.imports  = None
        self.service  = ''
        self.classdef = ''
        self.initdef  = ''
        self.location = ''
        self.methods  = []
        self.action_in = {}
        self.action_out = {}

    def getMethodName(self, method):
        return '%s_%s' %(self.method_prefix, method)

    def getStubName(self):
        return '%s_server' % self.service

    def fromWsdl(self, ws):
        wsm = ZSI.wsdl2python.WriteServiceModule(ws)
        self.service = wsm.get_module_names()[1]
        wsm = None
        ws = ZSI.wsdlInterface.ZSIWsdlAdapter(ws)
        self.setImports()
        for service in ws.getServicesList():
            for port in service.getPortList():
                # fetch the service location
                for e in port.getExtensions():
                    soapAddress = None
                    if isinstance(e, ZSI.wsdlInterface.ZSISoapAddressAdapter):
                        soapAddress = e
                    if soapAddress:
                        self.location = soapAddress.getLocation()

                # generate methods
                for op in port.getBinding().getPortType().getOperationList():
                    self.generateMethod(op, port)

        self.setClassDef(ws)
        self.setInitDef()

    def write(self, fd=sys.stdout):
        module_contents = \
            string.join([self.imports, self.classdef, self.initdef, string.join(self.methods, '\n')], '\n') + '\n'
        fd.write(module_contents)

    def setImports(self):
        i  = 'from %s.%s import *' %(self.stub_class_dir, self.service)
        i += '\nfrom %s import %s' %(self.base_class_mod, self.base_class)
        self.imports = i

    def setClassDef(self, ws):
        c  = '\nclass %s(%s):' %(ws.getName(), self.base_class)

        # Set up soapAction dict
        sa = '\n%ssoapAction = {' % ServerCallbackDescription.ID1
        for method,value in self.action_in.items():
            sa += "\n%s'%s': '%s'," % (ServerCallbackDescription.ID2, value, self.getMethodName(method))
        sa += '\n%s}' % ServerCallbackDescription.ID2

        # Set up wsAction dict
        wsa = '\n%swsAction = {' % ServerCallbackDescription.ID1
        for method,value in self.action_out.items():
            wsa += "\n%s'%s': '%s'," % (ServerCallbackDescription.ID2, self.getMethodName(method), value)
        wsa += '\n%s}' % ServerCallbackDescription.ID2

        self.classdef = c + sa + wsa

    def setInitDef(self):
        uri = urlparse.urlparse( self.location )[2]
        d  = "\n%sdef __init__(self, post='%s', **kw):" % (ServerCallbackDescription.ID1, uri)
        d += '\n%s%s.__init__(self, post)' %(ServerCallbackDescription.ID2, self.base_class)
        self.initdef  = d

    def generateMethod(self, op, port):
        # generate wsActions while we're here
        operation_name = op.getName()
        # Grab portType from Adapter
        portType = port.getBinding().getPortType()._portType
        operation = portType.operations.get(operation_name)
        wsaction_in = wsaction_out = None
        if operation.input is not None:
            wsaction_in = operation.getInputAction()
        if operation.output is not None:
            wsaction_out = operation.getOutputAction()
        soap_operation = port.getBinding().getOperationDict().get(operation_name).getSoapOperation()
        if ALLOW_SLOPPY_WSDL is False:
            soap_operation = port.getBinding().getOperationDict().get(operation_name).getSoapOperation()
            soapaction = soap_operation.getAction()
            if soapaction and wsaction_in and soapaction != wsaction_in:
                raise RuntimeError, 'Bad WSDL operation(%s), soapAction(%s) != wsAction(%s)'\
                    %(operation_name, soapaction, wsaction_in)

        self.action_in[operation_name] = wsaction_in
        self.action_out[operation_name] = wsaction_out
 
        # now take care of the method
        inMessage = op.getInput().getMessage()
        if inMessage is not None:
            o  = '\n%sdef %s(self, ps, address):' % (ServerCallbackDescription.ID1, self.getMethodName(op.getName()))
            o += '\n%s# input vals in request object' % ServerCallbackDescription.ID2
            o += '\n%sself.request = ps.Parse(%s())' \
                      % ( ServerCallbackDescription.ID2, op.getInput().getMessage().getName() )
        else:
            o  = '\n%sdef %s(self, ps):' % (ServerCallbackDescription.ID1, self.getMethodName(op.getName()))
            o += '\n%s# NO input' % ServerCallbackDescription.ID2
        
        outMessage = op.getOutput().getMessage()
        if outMessage is not None:
            o += '\n\n%s# assign return values to response object' % ServerCallbackDescription.ID2
            o += '\n%sresponse = %s()' % ( ServerCallbackDescription.ID2, outMessage.getName() )
            o += '\n\n%sreturn response' % ServerCallbackDescription.ID2
        else:
            o += '\n\n%s# NO output' % ServerCallbackDescription.ID2
            o += '\n\n%sreturn None' % ServerCallbackDescription.ID2

        self.methods.append(o)


#def run_happydoc():
#    doc = ["happydoc", "-d", "./doc", "pyGridWare", "formatter_levelOneHeadingBackgroundColor='#CCCCCC'", "formatter_levelOneHeadingForegroundColor='000000'", "formatter_levelTwoHeadingBackgroundColor='#CCCCCC'", "formatter_levelTwoHeadingForegroundColor='000000'", "./pyGridWare"]
#    spawn(doc, verbose=1)
#    return

class PyGridWareDist(Distribution): 
    global_options = Distribution.global_options[:]
    global_options.append(('regenerate', None, "regenerate service files"))

    def __init__ (self, attrs=None):
        # XXX Hook to run generator, wish there was a cleaner way??
        global_args = attrs.get('script_args', [])
        for arg in global_args:
            if arg == '--regenerate':
                RegenerateModules(services=True, stubs=True)
        Distribution.__init__(self, attrs)


def main():
    setup(name="pyGridWare", 
          version="0.0.2", 
          packages=["pyGridWare", "pyGridWare.addressing", "pyGridWare.security", "pyGridWare.security.gss",
          "pyGridWare.generated", "pyGridWare.generated.types", "pyGridWare.generated.stubs",
          "pyGridWare.generated.services", "pyGridWare.generated.properties", "pyGridWare.properties",
          "pyGridWare.resource", "pyGridWare.lifetime", "pyGridWare.resource", "pyGridWare.notification", 
          "pyGridWare.generated.resource", "pyGridWare.services", "pyGridWare.wsrf", "pyGridWare.utility"],
          description="Python WS-Resource Framework and GSS Security",
	  author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",
	  url="http://www-itg.lbl.gov/gtg/pyGridWare/",
          distclass=PyGridWareDist,
    )

if __name__ == '__main__': 
    main()
