#! /usr/bin/env python
# $Header: /cvsroot/pywebsvcs/zsi/ZSI/TC.py,v 1.32.2.21 2004/10/29 03:34:49 boverhof Exp $
'''General typecodes.
'''

from ZSI import _copyright, _children, _child_elements, \
    _floattypes, _stringtypes, _seqtypes, _find_attr, _find_attrNS, \
    _find_arraytype, _find_default_namespace, _find_href, _find_encstyle, \
    _textprotect, _textunprotect, _resolve_prefix, _find_xsi_attr, _find_type, \
    _find_xmlns_prefix, _get_element_nsuri_name, _Node, EvaluateException, \
    _valid_encoding

from ZSI.wstools import Base
from ZSI.wstools.Namespaces import SCHEMA, SOAP
from ZSI.wstools.Utility import SplitQName
from ZSI.compat import Canonicalize

import re, types, time

from base64 import decodestring as b64decode, encodestring as b64encode
from urllib import unquote as urldecode, quote as urlencode
from binascii import unhexlify as hexdecode, hexlify as hexencode


_is_xsd_or_soap_ns = lambda ns: ns in [
                        SCHEMA.XSD3, SOAP.ENC, SCHEMA.XSD1, SCHEMA.XSD2, ]
_find_nil = lambda E: _find_xsi_attr(E, "null") or _find_xsi_attr(E, "nil")

_is_element_declaration_typecode = lambda pyclass: hasattr(pyclass, 'literal')
_is_type_definition_typecode = lambda pyclass: hasattr(pyclass, 'type') and not hasattr(pyclass, 'literal') 


def _get_xsitype(pyclass):
    if hasattr(pyclass,'type') and type(pyclass.type) in _seqtypes:
        return pyclass.type
    elif hasattr(pyclass,'type') and hasattr(pyclass, 'schema'):
        return (pyclass.schema, pyclass.type)

    return (None,None)

def _get_object_id(pyobj):
    return '%x' % id(pyobj)

def _add_namespace_container(pyclass):
    raise DeprecationWarning, 'This method is deprecated'

def _get_type_definition(namespaceURI, name):
    return MetaClass.getTypeDefinition(namespaceURI, name)

def _has_namespaceURI(namespaceURI):
    raise DeprecationWarning, 'This method is deprecated'

def _get_global_element_declaration(namespaceURI, name):
    return MetaClass.getElementDeclaration(namespaceURI, name)

def _get_substitute_element(elt, what):
    raise NotImplementedError, 'Not implemented'

def _has_type_definition(namespaceURI, name):
    return MetaClass.getTypeDefinition(namespaceURI, name) is not None



class MetaClass(type):
    '''Store all types/elements in dictionary, when hit
    already defined class dont create a new one just
    give back reference.  Thus import order determines
    which class is loaded.
    '''
    types = {}
    elements = {}

    def __new__(cls,classname,bases,classdict):
        '''If classdict has literal and schema register it as a
        element declaration, else if has type and schema register
        it as a type definition.
        '''
        if classdict.get('schema',None) is not None and classdict.has_key('literal'):
            key = (classdict['schema'], classdict['literal'])
            if MetaClass.elements.has_key(key) is False:
                MetaClass.elements[key] = type.__new__(cls,classname,bases,classdict)
            return MetaClass.elements[key]
        elif classdict.get('schema',None) is not None and classdict.has_key('type'):
            key = classdict['type']
            if MetaClass.types.has_key(key) is False:
                MetaClass.types[key] = type.__new__(cls,classname,bases,classdict)
            return MetaClass.types[key]
        return type.__new__(cls,classname,bases,classdict)

    def getTypeDefinition(cls, namespaceURI, name):
        '''Grab a type definition
        namespaceURI -- 
        name -- 
        '''
        return cls.types.get((namespaceURI, name), None)
    getTypeDefinition = classmethod(getTypeDefinition)

    def getElementDeclaration(cls, namespaceURI, name):
        '''Grab an element declaration
        namespaceURI -- 
        name -- 
        '''
        return cls.elements.get((namespaceURI, name), None)
    getElementDeclaration = classmethod(getElementDeclaration)


class TypeCode(Base):
    '''The parent class for all parseable SOAP types.
    Class data:
        typechecks -- do init-time type checking if non-zero
    Class data subclasses may define:
        tag -- global element declaration
        type -- global type definition
        parselist -- list of valid SOAP types for this class, as
            (uri,name) tuples, where a uri of None means "all the XML
            Schema namespaces"
        errorlist -- parselist in a human-readable form; will be
            generated if/when needed
        seriallist -- list of Python types or user-defined classes
            that this typecode can serialize.
    '''
    tag = None
    type = (None,None)
    typechecks = True
    attribute_typecode_dict = None

    def __init__(self, pname=None, oname=None, aname=None, minOccurs=1,
    maxOccurs=1, nillable=False, typed=True, unique=True, attrs_aname='_attrs',
    **kw):
        '''Baseclass initialization.
        Instance data (and usually keyword arg)
            pname -- the parameter name (localname).
            nspname -- the namespace for the parameter;
                None to ignore the namespace
            typed -- output xsi:type attribute
            unique -- data item is not aliased (no href/id needed)
            minOccurs -- min occurances
            maxOccurs -- max occurances
            nillable -- is item nillable?
            attrs_aname -- This is variable name to dictionary of attributes
        '''
        Base.__init__(self)
        if type(pname) in _seqtypes:
            self.nspname, self.pname = pname
        else:
            self.nspname, self.pname = None, pname

        if self.pname:
            i = str(self.pname).find(':')
            if i > -1: self.pname = self.pname[i + 1:]

        self.aname = aname or self.pname
        self.minOccurs = minOccurs
        self.maxOccurs = maxOccurs
        self.nillable = nillable
        self.typed = typed
        self.unique = unique
        self.attrs_aname = attrs_aname
        #if kw.has_key('default'): self.default = kw['default']
        #if kw.has_key('rpc'): self.rpc = kw['rpc']

    def parse(self, elt, ps):
        '''elt -- the DOM element being parsed
        ps -- the ParsedSoap object.
        '''
        raise EvaluateException("Unimplemented evaluation", ps.Backtrace(elt))

    def serialize(self, elt, sw, pyobj, name=None, orig=None, **kw):
        '''elt -- the DOM element being parsed
           sw -- soapWriter object
           pyobj -- python object to serialize

        '''
        raise EvaluateException("Unimplemented evaluation", sw.Backtrace(elt))

    def SimpleHREF(self, elt, ps, tag):
        '''Simple HREF for non-string and non-struct and non-array.
        '''
        if len(_children(elt)): return elt
        href = _find_href(elt)
        if not href:
            if self.minOccurs is 0: return None
            raise EvaluateException('Non-optional ' + tag + ' missing',
                    ps.Backtrace(elt))
        return ps.FindLocalHREF(href, elt, 0)

    def get_parse_and_errorlist(self):
        """Get the parselist and human-readable version, errorlist is returned,
        because it is used in error messages.
        """
        d = self.__class__.__dict__
        parselist = d.get('parselist')
        errorlist = d.get('errorlist')
        if parselist and not errorlist:
            errorlist = []
            for t in parselist:
                if t[1] not in errorlist: errorlist.append(t[1])
            errorlist = ' or '.join(errorlist)
            d['errorlist'] = errorlist
        return (parselist, errorlist)

    def checkname(self, elt, ps):
        '''See if the name and type of the "elt" element is what we're
        looking for.   Return the element's type.
        '''

        parselist,errorlist = self.get_parse_and_errorlist()
        ns, name = _get_element_nsuri_name(elt)
        if ns == SOAP.ENC:
            # Element is in SOAP namespace, so the name is a type.
            if parselist and \
            (None, name) not in parselist and (ns, name) not in parselist:
                raise EvaluateException(
                'Element mismatch (got %s wanted %s) (SOAP encoding namespace)' % \
                        (name, errorlist), ps.Backtrace(elt))
            return (ns, name)

        # Not a type, check name matches.
        if self.nspname and ns != self.nspname:
            raise EvaluateException('Element NS mismatch (got %s wanted %s)' % \
                (ns, self.nspname), ps.Backtrace(elt))

        if self.pname and name != self.pname:
            raise EvaluateException('Element Name mismatch (got %s wanted %s)' % \
                (name, self.pname), ps.Backtrace(elt))
        return self.checktype(elt, ps)

    def checktype(self, elt, ps):
        '''See if the type of the "elt" element is what we're looking for.
        Return the element's type.
        '''
        type = _find_type(elt)
        if type is None or type == "":
            return (None,None)

        # Parse the QNAME.
        list = type.split(':')
        if len(list) == 1: 
            list = [None,list[0]] 
        elif len(list) != 2:
            raise EvaluateException('Malformed type attribute (not two colons)',
                    ps.Backtrace(elt))

        uri = ps.GetElementNSdict(elt).get(list[0])
        if uri is None:
            raise EvaluateException('Malformed type attribute (bad NS)',
                    ps.Backtrace(elt))
        type = list[1]
        parselist,errorlist = self.get_parse_and_errorlist()
        if not parselist or \
        (uri,type) in parselist or \
        (_is_xsd_or_soap_ns(uri) and (None,type) in parselist):
            return (uri,type)
        raise EvaluateException(
                'Type mismatch (%s namespace) (got %s wanted %s)' % \
                (uri, type, errorlist), ps.Backtrace(elt))

    def name_match(self, elt):
        '''Simple boolean test to see if we match the element name.
        '''
        return self.pname == elt.localName and \
                    self.nspname in [None, elt.namespaceURI]

    def nilled(self, elt, ps):
        '''Is the element NIL, and is that okay?
        '''
        if _find_nil(elt) not in [ "true",  "1"]: return False
        if self.nillable is False:
            raise EvaluateException('Non-nillable element is NIL',
                    ps.Backtrace(elt))
        return True

    def simple_value(self, elt, ps):
        '''Get the value of the simple content of this element.
        '''
        if not _valid_encoding(elt):
            raise EvaluateException('Invalid encoding', ps.Backtrace(elt))
        c = _children(elt)
        if len(c) == 0:
            raise EvaluateException('Value missing', ps.Backtrace(elt))
        for c_elt in c:
            if c_elt.nodeType == _Node.ELEMENT_NODE:
                raise EvaluateException('Sub-elements in value',
                    ps.Backtrace(c_elt))

        # It *seems* to be consensus that ignoring comments and
        # concatenating the text nodes is the right thing to do.
        return ''.join([E.nodeValue for E in c
                if E.nodeType 
                in [ _Node.TEXT_NODE, _Node.CDATA_SECTION_NODE ]])

    def parse_attributes(self, elt):
        '''find all attributes specified in the attribute_typecode_dict in
        current element tag, if an attribute is found set it in the 
        self.attributes dictionary.  Default to putting in String.
        '''
        attributes = {}
        if self.attribute_typecode_dict is None: return
        for attr,what in self.attribute_typecode_dict.items():
            if type(attr) in _seqtypes: 
                namespaceURI,localName = attr
                value = _find_attrNS(elt, namespaceURI, localName)
                self.logger.debug("Parsed Attribute (%s,%s) -- %s", namespaceURI, localName, value)
            else:
                value = _find_attr(elt, attr)
                self.logger.debug("Parsed Attribute (%s) -- %s", attr, value)
            # For Now just set it w/o any type interpretation.
            if value is None: continue
            attributes[attr] = value
        return attributes

    def set_attributes(self, el, pyobj):
        '''Instance data attributes contains a dictionary 
        of keys (namespaceURI,localName) and attribute values.
        These values can be self-describing (typecode), or use
        attribute_typecode_dict to determine serialization.
            el -- MessageInterface instance representing the element
                declaration.
        '''
        if not hasattr(pyobj, self.attrs_aname):
            return
        if not isinstance(getattr(pyobj, self.attrs_aname), dict):
            raise TypeError, 'pyobj.%s must be a dictionary of names and values'
        for attr, value in getattr(pyobj, self.attrs_aname).items():
            namespaceURI,localName = None, attr
            if type(attr) in _seqtypes:
                namespaceURI, localName = attr
            if getattr(self, 'attribute_typecode_dict', None) is not None:
                what = self.attribute_typecode_dict.get(attr)
                if what is None and namespaceURI is None:
                    what = self.attribute_typecode_dict.get(localName)
            else:
                what = Any()
            if hasattr(value, 'typecode'):
                if not isinstance(value.typecode,what):
                    raise EvaluateException, 'self-describing attribute must subclass %s' %what.__class__
                value = value.typecode.get_formatted_content(value)
            elif what is None:
                raise EvaluateException, 'Missing typecode for attribute(%s,%s) of element (%s,%s)'\
                    %(namespaceURI, localName, self.nspname, self.pname)
            else:
                self.logger.debug("attribute create -- %s", value)
                value = what.get_formatted_content(value)
            el.setAttributeNS(namespaceURI, localName, value)

    def set_attribute_xsi_type(self, el, **kw):
        '''if typed, set the xsi:type attribute 
        '''
        if kw.get('typed', self.typed):
            namespaceURI,typeName = kw.get('type', _get_xsitype(self))
            if namespaceURI and typeName:
                self.logger.debug("attribute: (%s, %s)", namespaceURI, typeName)
                el.setAttributeType(namespaceURI, typeName)

    def set_attribute_href(self, el, objid):
        '''set href attribute
        '''
        el.setAttributeNS(None, 'href', "#%s" %objid)

    def set_attribute_id(self, el, objid):
        '''set id attribute
        '''
        if not self.unique:
            el.setAttributeNS(None, 'id', "%s" %objid)

    def get_name(self, name, objid):
        n = name or self.pname or ('E' + objid)
        return n

    def has_attributes(self):
        '''Return True if Attributes are declared outside
        the scope of SOAP ('root', 'id', 'href'), and some
        attributes automatically handled (xmlns, xsi:type).
        '''
        if self.attribute_typecode_dict is None: return False
        return len(self.attribute_typecode_dict) > 0


class SimpleType(TypeCode):
    '''SimpleType -- consist exclusively of a tag, attributes, and a value
    '''

    def get_formatted_content(self, pyobj):
        raise NotImplementedError, 'method get_formatted_content is not implemented'

    def serialize(self, elt, sw, pyobj, name=None, orig=None, **kw):
        '''Handles the start and end tags, and attributes.  callout
        to get_formatted_content to get the textNode value.
            sw --
            pyobj -- processed content.
        '''
        objid = _get_object_id(pyobj)
        n = name or self.pname or ('E' + objid)

        #Nillable
        el = elt.createAppendElement(self.nspname, n)
        if self.nillable and pyobj is None:
            el.setAttributeNS(SCHEMA.XSI3, 'nil', '1')
            return None

        #other attributes
        self.set_attributes(el, pyobj)

        #href attribute
        if type(pyobj) in _seqtypes:
            self.set_attribute_href(el, pyobj[0])
            return None
        unique = self.unique or kw.get('unique', False)
        if not unique and sw.Known(orig or pyobj):
            self.set_attribute_href(el, objid)
            return None

        #xsi:type attribute 
        self.set_attribute_xsi_type(el, **kw)

        #id attribute
        if not self.unique:
            self.set_attribute_id(el, objid)

        #Content, <empty tag/>c
        if pyobj is not None:
            textNode = self.get_formatted_content(pyobj)
            if type(textNode) not in _stringtypes:
                raise TypeError, 'pyobj must be a formatted string'
            el.createAppendTextNode(textNode)
        return el


#XXX NOT FIXED YET
class Any(TypeCode):
    '''When the type isn't defined in the schema, but must be specified
    in the incoming operation.
        parsemap -- a type to class mapping (updated by descendants), for
                parsing
        serialmap -- same, for (outgoing) serialization
    '''
    parsemap, serialmap = {}, {}

    def __init__(self, pname=None, aslist=0, **kw):
        TypeCode.__init__(self, pname, **kw)
        self.aslist = aslist
        # If not derived, and optional isn't set, make us optional
        # so that None can be parsed.
        if self.__class__ == Any and not kw.has_key('optional'):
            self.optional = 1

    def listify(self, v):
        if self.aslist: return [ v[k] for k in v.keys() ]
        return v

    def parse_into_dict_or_list(self, elt, ps):
        c = _child_elements(elt)
        count = len(c)
        v = {}
        if count == 0:
            href = _find_href(elt)
            if not href: return v
            elt = ps.FindLocalHREF(href, elt)
            self.checktype(elt, ps)
            c = _child_elements(elt)
            count = len(c)
            if count == 0: return self.listify(v)
        if self.nilled(elt, ps): return None
        for c_elt in c:
            v[str(c_elt.nodeName)] = self.parse(c_elt, ps)
        return self.listify(v)

    def parse(self, elt, ps):
        (ns,type) = self.checkname(elt, ps)
        if not type and self.nilled(elt, ps): return None
        if len(_children(elt)) == 0:
            href = _find_href(elt)
            if not href:
                if self.optional: return None
                raise EvaluateException('Non-optional Any missing',
                        ps.Backtrace(elt))
            elt = ps.FindLocalHREF(href, elt)
            (ns,type) = self.checktype(elt, ps)
        if not type and elt.namespaceURI == SOAP.ENC:
            ns,type = SOAP.ENC, elt.localName
        if not type or (ns,type) == (SOAP.ENC,'Array'):
            if self.aslist or _find_arraytype(elt):
                return [ Any(aslist=self.aslist).parse(e, ps)
                            for e in _child_elements(elt) ]
            if len(_child_elements(elt)) == 0:
                raise EvaluateException("Any cannot parse untyped element",
                        ps.Backtrace(elt))
            return self.parse_into_dict_or_list(elt, ps)
        parser = Any.parsemap.get((ns,type))
        if not parser and _is_xsd_or_soap_ns(ns):
            parser = Any.parsemap.get((None,type))
        if not parser:
            raise EvaluateException('''Any can't parse element''',
                    ps.Backtrace(elt))
        return parser.parse(elt, ps)

    def get_formatted_content(self, pyobj):
        tc = type(pyobj)
        if tc == types.InstanceType:
            tc = pyobj.__class__
            if hasattr(pyobj, 'typecode'):
                serializer = pyobj.typecode.serialmap.get(tc)
            else:
                serializer = Any.serialmap.get(tc)
            if not serializer:
                tc = (types.ClassType, pyobj.__class__.__name__)
                serializer = Any.serialmap.get(tc)
        else:
            serializer = Any.serialmap.get(tc)
            if not serializer and isinstance(pyobj, time.struct_time):
                from ZSI.TCtimes import gDateTime
                serializer = gDateTime()
        if serializer:
            return serializer.get_formatted_content(pyobj)
        raise EvaluateException, 'Failed to find serializer for pyobj %s' %pyobj

    def serialize(self, elt, sw, pyobj, name=None, rpc=None, **kw):
        if hasattr(pyobj, 'typecode'):
            pyobj.typecode.serialize(elt, sw, pyobj, **kw)
            return

        #n = name or self.oname or rpc or 'E%x' % id(pyobj)
        objid = _get_object_id(pyobj)
        n = self.get_name(name, objid)

        kw['name'] = n
        tc = type(pyobj)
        self.logger.debug('Any serialize -- %s', tc)
        if tc == types.DictType or self.aslist:
            """
            #if rpc != None:
            #    print >>sw, '<%s>' % rpc
            #else:
            #    if type(pyobj) != types.InstanceType:
            #        print >>sw, '<%s>' % n
            if self.aslist:
                for val in pyobj:
                    Any().serialize(elt, sw, val)
            else:
                for key,val in pyobj.items():
                    Any(pname=key).serialize(elt, sw, val)
            #if rpc != None:
            #    print >>sw, '</%s>' % rpc
            #else:
            #    if type(pyobj) != types.InstanceType:
            #        print >>sw, '</%s>' % n
            return
            """
            raise EvaluateException, 'Not implemented yet'

        if tc in _seqtypes:
            raise EvaluateException, 'Not implemented yet'

        if tc == types.InstanceType:
            tc = pyobj.__class__
            if hasattr(pyobj, 'typecode'):
                serializer = pyobj.typecode.serialmap.get(tc)
            else:
                serializer = Any.serialmap.get(tc)
            if not serializer:
                tc = (types.ClassType, pyobj.__class__.__name__)
                serializer = Any.serialmap.get(tc)
        else:
            serializer = Any.serialmap.get(tc)
            if not serializer and isinstance(pyobj, time.struct_time):
                from ZSI.TCtimes import gDateTime
                serializer = gDateTime()
        if not serializer:
            # Last-chance; serialize instances as dictionary
            if type(pyobj) != types.InstanceType:
                raise EvaluateException('''Any can't serialize ''' + \
                        repr(pyobj))
            self.serialize(elt, sw, pyobj.__dict__, **kw)
        else:
            # Try to make the element name self-describing
            tag = getattr(serializer, 'tag', None)
            if self.pname is not None:
                serializer.nspname = self.nspname
                serializer.pname = self.pname
                kw = {'typed':0}
            elif tag:
                if tag.find(':') == -1: tag = 'SOAP-ENC:' + tag
                kw['name'] = kw['oname'] = tag
                kw['typed'] = 0
            serializer.serialize(elt, sw, pyobj, **kw)
            # Reset TypeCode
            serializer.nspname = None
            serializer.pname = None


def RegisterType(C, clobber=0, *args, **keywords):
    instance = apply(C, args, keywords)
    for t in C.__dict__.get('parselist', []):
        prev = Any.parsemap.get(t)
        if prev:
            if prev.__class__ == C: continue
            if not clobber:
                raise TypeError(
                    str(C) + ' duplicating parse registration for ' + str(t))
        Any.parsemap[t] = instance
    for t in C.__dict__.get('seriallist', []):
        ti = type(t)
        if ti in [ types.TypeType, types.ClassType]:
            key = t
        elif ti in _stringtypes:
            key = (types.ClassType, t)
        else:
            raise TypeError(str(t) + ' is not a class name')
        prev = Any.serialmap.get(key)
        if prev:
            if prev.__class__ == C: continue
            if not clobber:
                raise TypeError(
                    str(C) + ' duplicating serial registration for ' + str(t))
        Any.serialmap[key] = instance


class String(SimpleType):
    '''A string type.
    '''
    parselist = [ (None,'string') ]
    seriallist = [ types.StringType, types.UnicodeType ]
    type = (SCHEMA.XSD3, 'string')

    def __init__(self, pname=None, strip=1, textprotect=1, **kw):
        TypeCode.__init__(self, pname, **kw)
        if kw.has_key('resolver'): self.resolver = kw['resolver']
        self.strip = strip
        self.textprotect = textprotect

    def parse(self, elt, ps):
        self.checkname(elt, ps)
        if len(_children(elt)) == 0:
            href = _find_href(elt)
            if not href:
                if self.nilled(elt, ps) is False:
                    # No content, no HREF, not NIL:  empty string
                    return ""
                # No content, no HREF, and is NIL...
                if self.minOccurs == 0 or self.nillable is True: 
                    return None
                raise EvaluateException('Non-optional string missing',
                        ps.Backtrace(elt))
            if href[0] != '#':
                return ps.ResolveHREF(href, self)
            elt = ps.FindLocalHREF(href, elt)
            self.checktype(elt, ps)
        if self.nilled(elt, ps): return None
        if len(_children(elt)) == 0: return ''
        v = self.simple_value(elt, ps)
        if self.strip: v = v.strip()
        if self.textprotect: v = _textunprotect(v)
        return v

    def get_formatted_content(self, pyobj):
        if type(pyobj) == types.UnicodeType: pyobj = pyobj.encode('utf-8')
        if self.textprotect: pyobj = _textprotect(pyobj)
        return pyobj


class URI(String):
    '''A URI.
    '''
    parselist = [ (None,'anyURI') ]
    type = (SCHEMA.XSD3, 'anyURI')

    def parse(self, elt, ps):
        val = String.parse(self, elt, ps)
        #return urldecode(val)
        return val

    def get_formatted_content(self, pyobj):
        #pyobj = urlencode(pyobj)
        return String.get_formatted_content(self, pyobj)



class QName(String):
    '''A QName type
    '''
    parselist = [ (None,'QName') ]
    type = (SCHEMA.XSD3, 'QName')

    def __init__(self, pname=None, strip=1, textprotect=1, **kw):
        String.__init__(self, pname, strip, textprotect, **kw)
        self.prefix = None

    def get_formatted_content(self, pyobj):
        value = pyobj
        if isinstance(pyobj, tuple):
            namespaceURI,localName = pyobj
            value = "%s:%s" %(self.prefix,localName)
        return String.get_formatted_content(self, value)

    def parse(self, elt, ps):
        val = String.parse(self, elt, ps)
        prefix,localName = SplitQName(val)
        nsdict = ps.GetElementNSdict(elt)
        try:
            namespaceURI = nsdict[prefix]
        except KeyError, ex:
            raise EvaluateException, 'cannot resolve prefix(%s)'%prefix
        return (namespaceURI,localName)

    def serialize(self, elt, sw, pyobj, name=None, orig=None, **kw):
        '''pyobj -- qualifiedName or (namespaceURI,elementName) tuple
        '''
        if isinstance(pyobj, tuple):
            namespaceURI,localName = pyobj
            self.prefix = elt.getPrefix(namespaceURI)
        else:
            try:
                namespaceURI = elt.findNamespaceURI(pyobj)
            except Exception, ex:
                raise EvaluateException, 'QName (%s) has undeclared prefix' %pyobj
        String.serialize(self, elt, sw, pyobj, name, orig, **kw)


class Token(String):
    '''an xsd:token type
    '''
    parselist = [ (None, 'token') ]
    type = (SCHEMA.XSD3, 'token')


class Base64String(String):
    '''A Base64 encoded string.
    '''
    parselist = [ (None,'base64Binary'), (SOAP.ENC, 'base64') ]
    type = (SOAP.ENC, 'base64')

    def parse(self, elt, ps):
        val = String.parse(self, elt, ps)
        return b64decode(val.replace(' ', '').replace('\n','').replace('\r',''))

    def get_formatted_content(self, pyobj):
        pyobj = '\n' + b64encode(pyobj)
        return String.get_formatted_content(self, pyobj)


class Base64Binary(String):
    parselist = [ (None,'base64Binary'), ]
    type = (SCHEMA.XSD3, 'base64Binary')

    def parse(self, elt, ps):
        val = String.parse(self, elt, ps)
        #return b64decode(val.replace(' ', '').replace('\n','').replace('\r',''))
        if val is None: 
            if self.nillable is True: return None
            raise EvaluateException, "Base64Binary (%s,%s) is not nillable" %(self.nspname,self.pname)
        return b64decode(val)

    def get_formatted_content(self, pyobj):
        pyobj = b64encode(pyobj).strip()
        #return String.get_formatted_content(self, pyobj)
        return pyobj

class HexBinaryString(String):
    '''Hex-encoded binary (yuk).
    '''
    parselist = [ (None,'hexBinary') ]
    type = (SCHEMA.XSD3, 'hexBinary')

    def parse(self, elt, ps):
        val = String.parse(self, elt, ps)
        return hexdecode(val)

    def get_formatted_content(self, pyobj):
        pyobj = hexencode(pyobj).upper()
        return String.get_formatted_content(self, pyobj)


class XMLString(String):
    '''A string that represents an XML document
    '''

    def __init__(self, pname=None, readerclass=None, **kw):
        String.__init__(self, pname, **kw)
        self.readerclass = readerclass

    def parse(self, elt, ps):
        if not self.readerclass:
            from xml.dom.ext.reader import PyExpat
            self.readerclass = PyExpat.Reader
        v = String.parse(self, elt, ps)
        return self.readerclass().fromString(v)

    def get_formatted_content(self, pyobj):
        pyobj = Canonicalize(pyobj)
        return String.get_formatted_content(self, pyobj)


class Enumeration(String):
    '''A string type, limited to a set of choices.
    '''

    def __init__(self, choices, pname=None, **kw):
        String.__init__(self, pname, **kw)
        t = type(choices)
        if t in _seqtypes:
            self.choices = tuple(choices)
        elif TypeCode.typechecks:
            raise TypeError(
                'Enumeration choices must be list or sequence, not ' + str(t))
        if TypeCode.typechecks:
            for c in self.choices:
                if type(c) not in _stringtypes:
                    raise TypeError(
                        'Enumeration choice ' + str(c) + ' is not a string')

    def parse(self, elt, ps):
        val = String.parse(self, elt, ps)
        if val not in self.choices:
            raise EvaluateException('Value not in enumeration list',
                    ps.Backtrace(elt))
        return val


# This is outside the Integer class purely for code esthetics.
_ignored = []

class Integer(SimpleType):
    '''Common handling for all integers.
    '''

    ranges = {
        'unsignedByte':         (0, 255),
        'unsignedShort':        (0, 65535),
        'unsignedInt':          (0, 4294967295L),
        'unsignedLong':         (0, 18446744073709551615L),

        'byte':                 (-128, 127),
        'short':                (-32768, 32767),
        'int':                  (-2147483648L, 2147483647),
        'long':                 (-9223372036854775808L, 9223372036854775807L),

        'negativeInteger':      (_ignored, -1),
        'nonPositiveInteger':   (_ignored, 0),
        'nonNegativeInteger':   (0, _ignored),
        'positiveInteger':      (1, _ignored),

        'integer':              (_ignored, _ignored)
    }
    parselist = [ (None,k) for k in ranges.keys() ]
    seriallist = [ types.IntType, types.LongType ]

    def __init__(self, pname=None, format='%d', **kw):
        TypeCode.__init__(self, pname, **kw)
        self.format = format

    def parse(self, elt, ps):
        (ns,type) = self.checkname(elt, ps)
        if self.nilled(elt, ps): return None
        elt = self.SimpleHREF(elt, ps, 'integer')
        if not elt: return None

        if type is None:
           type = self.type[1] 
           #ns,type = _get_type_definition(*self.type)
        elif self.type[1] is not None and type != self.type[1]:
            raise EvaluateException('Integer type mismatch; ' \
                'got %s wanted %s' % (type,type[1]), ps.Backtrace(elt))
            
        
        #if self.nilled(elt, ps): return None
        v = self.simple_value(elt, ps)
        try:
            v = int(v)
        except:
            try:
                v = long(v)
            except:
                raise EvaluateException('Unparseable integer',
                    ps.Backtrace(elt))
        (rmin, rmax) = Integer.ranges.get(type, (_ignored, _ignored))
        if rmin != _ignored and v < rmin:
            raise EvaluateException('Underflow, less than ' + repr(rmin),
                    ps.Backtrace(elt))
        if rmax != _ignored and v > rmax:
            raise EvaluateException('Overflow, greater than ' + repr(rmax),
                    ps.Backtrace(elt))
        return v

    def get_formatted_content(self, pyobj):
        return self.format %pyobj



# See credits, below.
def _make_inf():
    x = 2.0
    x2 = x * x
    i = 0
    while i < 100 and x != x2:
        x = x2
        x2 = x * x
        i = i + 1
    if x != x2:
        raise ValueError("This machine's floats go on forever!")
    return x

# This is outside the Decimal class purely for code esthetics.
_magicnums = { }
try:
    _magicnums['INF'] = float('INF')
    _magicnums['-INF'] = float('-INF')
except:
    _magicnums['INF'] = _make_inf()
    _magicnums['-INF'] = -_magicnums['INF']

# The following comment and code was written by Tim Peters in
# article <001401be92d2$09dcb800$5fa02299@tim> in comp.lang.python,
# also available at the following URL:
# http://groups.google.com/groups?selm=001401be92d2%2409dcb800%245fa02299%40tim
# Thanks, Tim!

# NaN-testing.
#
# The usual method (x != x) doesn't work.
# Python forces all comparisons thru a 3-outcome cmp protocol; unordered
# isn't a possible outcome.  The float cmp outcome is essentially defined
# by this C expression (combining some cross-module implementation
# details, and where px and py are pointers to C double):
#   px == py ? 0 : *px < *py ? -1 : *px > *py ? 1 : 0
# Comparing x to itself thus always yields 0 by the first clause, and so
# x != x is never true.
# If px and py point to distinct NaN objects, a strange thing happens:
# 1. On scrupulous 754 implementations, *px < *py returns false, and so
#    does *px > *py.  Python therefore returns 0, i.e. "equal"!
# 2. On Pentium HW, an unordered outcome sets an otherwise-impossible
#    combination of condition codes, including both the "less than" and
#    "equal to" flags.  Microsoft C generates naive code that accepts
#    the "less than" flag at face value, and so the *px < *py clause
#    returns true, and Python returns -1, i.e. "not equal".
# So with a proper C 754 implementation Python returns the wrong result,
# and under MS's improper 754 implementation Python yields the right
# result -- both by accident.  It's unclear who should be shot <wink>.
#
# Anyway, the point of all that was to convince you it's tricky getting
# the right answer in a portable way!
def isnan(x):
    """x -> true iff x is a NaN."""
    # multiply by 1.0 to create a distinct object (x < x *always*
    # false in Python, due to object identity forcing equality)
    if x * 1.0 < x:
        # it's a NaN and this is MS C on a Pentium
        return 1
    # Else it's non-NaN, or NaN on a non-MS+Pentium combo.
    # If it's non-NaN, then x == 1.0 and x == 2.0 can't both be true,
    # so we return false.  If it is NaN, then assuming a good 754 C
    # implementation Python maps both unordered outcomes to true.
    return 1.0 == x and x == 2.0


class Decimal(SimpleType):
    '''Parent class for floating-point numbers.
    '''

    parselist = [ (None,'decimal'), (None,'float'), (None,'double') ]
    seriallist = _floattypes
    type = None
    ranges =  {
        'float': ( 7.0064923216240861E-46,
                        -3.4028234663852886E+38, 3.4028234663852886E+38 ),
        'double': ( 2.4703282292062327E-324,
                        -1.7976931348623158E+308, 1.7976931348623157E+308),
    }
    zeropat = re.compile('[1-9]')

    def __init__(self, pname=None, format='%f', **kw):
        TypeCode.__init__(self, pname, **kw)
        self.format = format

    def parse(self, elt, ps):
        (ns,type) = self.checkname(elt, ps)
        elt = self.SimpleHREF(elt, ps, 'floating-point')
        if not elt: return None
        tag = getattr(self.__class__, 'type')
        if tag:
            if type is None:
                type = tag
            elif tag != type:
                raise EvaluateException('Floating point type mismatch; ' \
                        'got %s wanted %s' % (type,tag), ps.Backtrace(elt))
        # Special value?
        if self.nilled(elt, ps): return None
        v = self.simple_value(elt, ps)
        m = _magicnums.get(v)
        if m: return m

        try:
            fp = float(v)
        except:
            raise EvaluateException('Unparseable floating point number',
                    ps.Backtrace(elt))
        if str(fp).lower() in [ 'inf', '-inf', 'nan', '-nan' ]:
            raise EvaluateException('Floating point number parsed as "' + \
                    str(fp) + '"', ps.Backtrace(elt))
        if fp == 0 and Decimal.zeropat.search(v):
            raise EvaluateException('Floating point number parsed as zero',
                    ps.Backtrace(elt))
        (rtiny, rneg, rpos) = Decimal.ranges.get(type, (None, None, None))
        if rneg and fp < 0 and fp < rneg:
            raise EvaluateException('Negative underflow', ps.Backtrace(elt))
        if rtiny and fp > 0 and fp < rtiny:
            raise EvaluateException('Positive underflow', ps.Backtrace(elt))
        if rpos and fp > 0 and fp > rpos:
            raise EvaluateException('Overflow', ps.Backtrace(elt))
        return fp

    def get_formatted_content(self, pyobj):
        if pyobj == _magicnums['INF']:
            return 'INF'
        elif pyobj == _magicnums['-INF']:
            return '-INF'
        elif isnan(pyobj):
            return 'NaN'
        else:
            return self.format %pyobj


class Boolean(SimpleType):
    '''A boolean.
    '''

    parselist = [ (None,'boolean') ]
    type = (SCHEMA.XSD3, 'boolean')

    def parse(self, elt, ps):
        self.checkname(elt, ps)
        elt = self.SimpleHREF(elt, ps, 'boolean')
        if not elt: return None
        if self.nilled(elt, ps): return None
        v = self.simple_value(elt, ps).lower()
        if v == 'false': return False
        if v == 'true': return True
        try:
            v = int(v)
        except:
            try:
                v = long(v)
            except:
                raise EvaluateException('Unparseable boolean',
                    ps.Backtrace(elt))
        if v: return True
        return False

    def get_formatted_content(self, pyobj):
        if pyobj: return 'true'
        return 'false'


#XXX NOT FIXED YET
class XML(TypeCode):
    '''Opaque XML which shouldn't be parsed.
        comments -- preserve comments
        inline -- don't href/id when serializing
        resolver -- object to resolve href's
        wrapped -- put a wrapper element around it
    '''

    # Clone returned data?
    copyit = 0

    def __init__(self, pname=None, comments=0, inline=0, wrapped=1, **kw):
        TypeCode.__init__(self, pname, **kw)
        self.comments = comments
        self.inline = inline
        if kw.has_key('resolver'): self.resolver = kw['resolver']
        self.wrapped = wrapped
        self.copyit = kw.get('copyit', XML.copyit)

    def parse(self, elt, ps):
        if not self.wrapped:
            return elt
        c = _child_elements(elt)
        if not c:
            href = _find_href(elt)
            if not href:
                if self.minOccurs == 0: return None
                raise EvaluateException('Embedded XML document missing',
                        ps.Backtrace(elt))
            if href[0] != '#':
                return ps.ResolveHREF(href, self)
            elt = ps.FindLocalHREF(href, elt)
            c = _child_elements(elt)
        if _find_encstyle(elt) != "":
            #raise EvaluateException('Embedded XML has unknown encodingStyle',
            #       ps.Backtrace(elt)
            pass
        if len(c) != 1:
            raise EvaluateException('Embedded XML has more than one child',
                    ps.Backtrace(elt))
        if self.copyit: return c[0].cloneNode(1)
        return c[0]

    def serialize(self, elt, sw, pyobj, name=None, unsuppressedPrefixes=[], **kw):
        if not self.wrapped:
            Canonicalize(pyobj, sw, unsuppressedPrefixes=unsuppressedPrefixes,
                comments=self.comments)
            return
        #objid = '%x' % id(pyobj)
        #n = name or self.oname or ('E' + objid)
        objid = _get_object_id(pyobj)
        n = name or self.pname or ('E' + objid)
        xmlelt = elt.createAppendElement(self.nspname, n)

        if type(pyobj) in _stringtypes:
            self.set_attributes(xmlelt, pyobj)
            self.set_attribute_href(xmlelt, objid)
        elif kw.get('inline', self.inline):
            self.cb(xmlelt, sw, pyobj, unsuppressedPrefixes)
        else:
            self.set_attributes(xmlelt, pyobj)
            self.set_attribute_href(xmlelt, objid)
            sw.AddCallback(self.cb, pyobj, unsuppressedPrefixes)

    def cb(self, elt, sw, pyobj, unsuppressedPrefixes=[]):
        if sw.Known(pyobj): return
        #objid = '%x' % id(pyobj)
        objid = _get_object_id(pyobj)
        n = self.pname or ('E' + objid)

        xmlelt = elt.createAppendElement(self.nspname, n)
        self.set_attribute_id(xmlelt, objid)
        xmlelt.setAttributeNS(SOAP.ENC, 'encodingStyle', '""')
        #print >>sw, '<%s SOAP-ENC:encodingStyle="" id="%s">' % (n, objid)
        Canonicalize(pyobj, sw, unsuppressedPrefixes=unsuppressedPrefixes,
            comments=self.comments)
        #print >>sw, '</%s>' % n


class AnyType(TypeCode):
    """XML Schema xsi:anyType type definition wildCard.
       class variables: 
          all -- specifies use of all namespaces.
          other -- specifies use of other namespaces
          type --
    """
    all = '#all'
    other = '#other'
    type = (SCHEMA.XSD3, 'anyType')

    def __init__(self, namespaces=['#all'],pname=None, 
    minOccurs=1, maxOccurs=1, strip=1, textprotect=1, **kw):
        TypeCode.__init__(self, pname=pname, minOccurs=minOccurs, maxOccurs=maxOccurs, **kw)
        self.namespaces = namespaces

    def serialize(self, elt, sw, pyobj, name=None, orig=None, **kw):
        nsuri,typeName = _get_xsitype(pyobj)
        if self.all not in self.namespaces and nsuri not in self.namespaces:
            raise EvaluateException, '<anyType> unsupported use of namespaces %s' %namespaces
        what = pyobj
        if hasattr(what, 'typecode'):
            what = pyobj.typecode
        elif not isinstance(what, TypeCode):
            if hasattr(what, 'typecode'):
                what = getattr(what, 'typecode')
            else:
                #May want to look thru containers and try to find a match
                #raise EvaluateException, '<anyType> pyobj must be self-describing.'
                what = Any(pname=(self.nspname,self.pname), aslist=False)
                what.serialize(elt, sw, pyobj, **kw)
                return
        what.nspname = self.nspname
        what.pname = self.pname
        what.serialize(elt, sw, pyobj, name=None, orig=None, **kw)

    def parse(self, elt, ps):
        #element name must be declared ..
        nspname,pname = _get_element_nsuri_name(elt)
        if nspname != self.nspname or pname != self.pname:
            raise EvaluateException, '<anyType> instance is (%s,%s) found (%s,%s)'\
                %(self.nspname,self.pname,nspname,pname)

        #locate xsi:type
        prefix, typeName = SplitQName(_find_type(elt))
        namespaceURI = _resolve_prefix(elt, prefix)
        pyclass = _get_type_definition(namespaceURI, typeName)
        if not pyclass:
            if _is_xsd_or_soap_ns(namespaceURI):
                #pyclass = BaseTypeInterpreter().get_typeclass(typeName,namespaceURI)
                pyclass = Any
            else:
                raise EvaluateException, '<anyType> cant find typecode for type (%s,%s)'\
                    %(namespaceURI,typeName)
        what = pyclass(pname=(self.nspname,self.pname))
        pyobj = what.parse(elt, ps)
        return pyobj


class AnyElement(AnyType):
    """XML Schema xsi:any element declaration wildCard.
       class variables: 
            tag -- global element declaration
    """
    tag = (SCHEMA.XSD3, 'any')

    def __init__(self, namespaces=['#all'],pname=None, 
    minOccurs=1, maxOccurs=1, strip=1, textprotect=1, processContents='strict',**kw):
        self.processContents = processContents
        AnyType.__init__(self, namespaces=namespaces,pname=pname,
            minOccurs=minOccurs, maxOccurs=maxOccurs, strip=strip, 
            textprotect=textprotect, **kw)

    def getProcessContents(self, processContents):
        return self._processContents
    def setProcessContents(self, processContents):
        if processContents not in ('lax', 'skip', 'strict'):
            raise EvaluateException, '<any> processContents(%s) is not understood.' %processContents
        self._processContents = processContents
    processContents = property(getProcessContents, setProcessContents, None, '<any> processContents')
       
    def serialize(self, elt, sw, pyobj, **kw):
        if isinstance(pyobj, TypeCode):
            what = pyobj
        elif hasattr(pyobj, 'typecode'):
            what = getattr(pyobj, 'typecode')
            #May want to look thru containers and try to find a match
        elif kw.has_key('pname'):
            what = Any(pname=(kw.get('nspname'),kw['pname']))
            #raise EvaluateException, '<anyType> pyobj must be self-describing.'
        else:
            what = Any()
        self.logger.debug('AnyElement.serialize with %s', what.__class__.__name__)
        what.serialize(elt, sw, pyobj, **kw)

    def parse(self, elt, ps):
        '''
        processContents -- 'lax' | 'skip' | 'strict', 'strict'
        1) if 'skip' check namespaces, and return the DOM node.
        2) if 'lax' look for declaration, or definition.  If
           not found return DOM node.
        3) if 'strict' get declaration, or raise.
        '''
        nspname,pname = _get_element_nsuri_name(elt)
        pyclass = _get_global_element_declaration(nspname, pname)
        if pyclass is None:
            # if self.processContents == 'strict': raise
            # Allow use of "<any>" element declarations w/ local element declarations
            prefix, typeName = SplitQName(_find_type(elt))
            if typeName:
                namespaceURI = _resolve_prefix(elt, prefix or 'xmlns')
                # First look thru user defined namespaces, if don't find
                # look for 'primitives'.
                pyclass = _get_type_definition(namespaceURI, typeName)
                if pyclass is None:
                    if not _is_xsd_or_soap_ns(namespaceURI):
                        raise EvaluateException, '<any> cant find typecode for type (%s,%s)'\
                            %(namespaceURI,typeName)
                    pyclass = Any
                what = pyclass(pname=(nspname,pname))
                pyobj = what.parse(elt, ps)
            else:
                # if self.processContents == 'lax' raise Exception
                # All else fails, not typed, attempt to use XML, String
                what = XML(pname=(nspname,pname))
                try:
                    pyobj = what.parse(elt, ps)
                except EvaluateException, ex:
                    self.logger.error("Give up, parse (%s,%s) as a String", what.nspname, what.pname)
                    # Try returning "elt"
                    pyobj = String(pname=(nspname,pname)).parse(elt, ps)

        else:
            what = pyclass()
            pyobj = what.parse(elt, ps)

        return pyobj


from TCnumbers import *
from TCtimes import *
from TCcompound import *
from TCapache import *

if __name__ == '__main__': print _copyright
