############################################################################
# Monte M. Goode, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################

# utility classes used by new generator - mostly 'sugar' classes
# that are actually imported by the generated code.  also includes
# utilities used by wsdl2python itself.

# $Id: utility.py,v 1.1.2.1 2004/10/21 20:27:13 tyger23 Exp $

import re
from ZSI import EvaluateException
from ZSI.TCcompound import Struct

# ---- generator utils

def _getLiteralFromTypeClassVar(classvar):
    """move this to type writer class?"""
    pass

def nonColonizedName_to_moduleName(name):
    """move back to main module?"""
    name = textProtect(name)
    return re.sub('\.', '_', name)

def textProtect(s):
    """move back to main module?"""
    if s:
        return re.sub('[-./:# ]', '_', s)
    else:
        return None


class NamespaceAliasDict:
    """a lookup table to store relevant namespaces and their aliases"""
    
    NSCOUNT = 1
    NSDICT = {}
    NSORDER = []

    def __init__(self):
        pass

    def add(self, ns):

        if not self.__class__.NSDICT.has_key(ns):
            self.__class__.NSDICT[ns] = ( self.namespace_to_moduleName(ns),
                                          'ns%s' % self.__class__.NSCOUNT )
            self.__class__.NSORDER.append(ns)
            self.__class__.NSCOUNT += 1
            
    def getModuleName(self, ns):

        if self.__class__.NSDICT.has_key(ns):
            return self.__class__.NSDICT[ns][0]
        else:
            raise WsdlGeneratorError, 'could not retrieve mod name for %s' % ns
        
    def getAlias(self, ns):

        if self.__class__.NSDICT.has_key(ns):
            return self.__class__.NSDICT[ns][1]
        else:
            self.dump()
            raise WsdlGeneratorError, 'could not retrieve alias for %s' % ns

    def getNSList(self):
        return self.__class__.NSORDER

    def namespace_to_moduleName(self, n):
        name = n

        if name.startswith('http://'):
            name = name[7:]

        name = textProtect(name)

        if name.endswith('_'):
            name = name[:-1]

        return name

    def dump(self):
        print self.__class__.NSDICT

    def reset(self):
        self.__class__.NSCOUNT = 1
        self.__class__.NSDICT = {}
        self.__class__.NSORDER = []




class StringWriter:
    """generator util"""
    def __init__(self, val=None):
        self.data = []
        if val:
            self.data.append(val)

    def set(self, val):
        if self.data:
            # in some cases the empty list reassignment fails, so....
            self.data = None
            self.data = []

        self.data.append(val)

    def write(self, val):
        self.data.append(val)

    def getvalue(self):
        if self.data:
            return ''.join(self.data)
        else:
            return ''

    def __iadd__(self, val):
        self.data.append(val)
        return self

    def __str__(self):
        return self.getvalue()


# ---- generated code utils

class MessageContainer:
    """generator util - used by address.py"""
    pass

class SchemaBase:
    """helper class for generated code"""

    schema = None
    
    def get_aname(name):
        '''Class method for constructing an aname from element
        declaration class information.
        '''
        if name is None:
            return None
        return '_%s' %name
    
    get_aname = staticmethod(get_aname)

    
    def get_nsuri(cls, **kw):
        namespaceURI = getattr(cls, 'schema')
        return namespaceURI
    
    get_nsuri = classmethod(get_nsuri)

    def get_variable(self, pyclass):
        '''Find instance of pyclass in ofwhat, and
        return the matching instance variable.
        '''
        if isinstance(self, Struct):
            for what in self.ofwhat:
                if isinstance(what, pyclass): break
            else:
                raise EvaluateException, '1'
        else:
            raise EvaluateException, '2'
        aname = getattr(what, 'aname', None)
        if aname is None:
            raise EvaluateException, '3'
        return getattr(self,aname)


class ElementDeclaration(SchemaBase):
    """helper class for generated code"""
    
    def get_pname(cls, **kw):
        '''Class method for retrieving element namespaceURI/name
        information, used to construct a pname. 
            Key Word:
                encoded -- encoded namespace (ignore for now).
        '''        
        namespaceURI = getattr(cls, 'schema')
        name = getattr(cls, 'literal')
        return (namespaceURI,name)
    
    get_pname = classmethod(get_pname)


class TypeDefinition(SchemaBase):
    """helper class for generated code"""
    pass


